#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ted Elhourani <ted@bigswitch.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: bcf_switch
author: "Ted (@tedelhourani)"
short_description: Create and remove a bcf switch.
description:
    - Create and remove a Big Cloud Fabric switch.
options:
  name:
    description:
     - The name of the switch.
    required: true
  fabric_role:
    description:
     - Fabric role of the switch.
    choices: ['spine', 'leaf']
    required: true
  leaf_group:
    description:
     - The leaf group of the switch if the switch is a leaf.
    required: false
  mac:
    description:
     - The MAC address of the switch.
    required: true
  state:
    description:
     - Whether the switch should be present or absent.
    default: present
    choices: ['present', 'absent']
  controller:
    description:
     - The controller IP address.
    required: true
  validate_certs:
    description:
      - If C(false), SSL certificates will not be validated. This should only be used
        on personally controlled devices using self-signed certificates.
    required: false
    default: true
    type: bool
  access_token:
    description:
     - Big Cloud Fabric access token. If this isn't set then the environment variable C(BIGSWITCH_ACCESS_TOKEN) is used.
'''


EXAMPLES = '''
- name: Bcf leaf switch
  community.network.bcf_switch:
    name: Rack1Leaf1
    fabric_role: leaf
    leaf_group: R1
    mac: 00:00:00:02:00:02
    controller: '{{ inventory_hostname }}'
    state: present
    validate_certs: false
'''


RETURN = ''' # '''

import os
import traceback

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.network.plugins.module_utils.network.bigswitch.bigswitch import Rest
from ansible.module_utils._text import to_native


def switch(module, check_mode):
    try:
        access_token = module.params['access_token'] or os.environ['BIGSWITCH_ACCESS_TOKEN']
    except KeyError as e:
        module.fail_json(msg='Unable to load %s' % e.message, exception=traceback.format_exc())

    name = module.params['name']
    fabric_role = module.params['fabric_role']
    leaf_group = module.params['leaf_group']
    dpid = '00:00:' + module.params['mac']
    state = module.params['state']
    controller = module.params['controller']

    rest = Rest(module,
                {'content-type': 'application/json', 'Cookie': 'session_cookie=' + access_token},
                'https://' + controller + ':8443/api/v1/data/controller/core')

    response = rest.get('switch-config', data={})
    if response.status_code != 200:
        module.fail_json(msg="failed to obtain existing switch config: {0}".format(response.json['description']))

    config_present = False
    for switch in response.json:
        if all((switch['name'] == name,
                switch['fabric-role'] == fabric_role,
                switch['dpid'] == dpid)):
            config_present = switch.get('leaf-group', None) == leaf_group
            if config_present:
                break

    if state in ('present') and config_present:
        module.exit_json(changed=False)

    if state in ('absent') and not config_present:
        module.exit_json(changed=False)

    if check_mode:
        module.exit_json(changed=True)

    if state in ('present'):
        data = {'name': name, 'fabric-role': fabric_role, 'leaf-group': leaf_group, 'dpid': dpid}
        response = rest.put('switch-config[name="%s"]' % name, data)
        if response.status_code == 204:
            module.exit_json(changed=True)
        else:
            module.fail_json(msg="error configuring switch '{0}': {1}".format(name, response.json['description']))

    if state in ('absent'):
        response = rest.delete('switch-config[name="%s"]' % name, data={})
        if response.status_code == 204:
            module.exit_json(changed=True)
        else:
            module.fail_json(msg="error deleting switch '{0}': {1}".format(name, response.json['description']))


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type='str', required=True),
            fabric_role=dict(choices=['spine', 'leaf'], required=True),
            leaf_group=dict(type='str', required=False),
            mac=dict(type='str', required=True),
            controller=dict(type='str', required=True),
            state=dict(choices=['present', 'absent'], default='present'),
            validate_certs=dict(type='bool', default='True'),
            access_token=dict(type='str', no_log=True)
        ),
        supports_check_mode=True,
    )

    try:
        switch(module, check_mode=module.check_mode)
    except Exception as e:
        module.fail_json(msg=to_native(e), exception=traceback.format_exc())


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright: (c) 2018, Abhijeet Kasurde <akasurde@redhat.com>
# Copyright: (c) 2018, Christian Kotte <christian.kotte@gmx.de>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: vmware_local_user_info
short_description: Gather info about users on the given ESXi host
description:
    - This module can be used to gather information about users present on the given ESXi host system in VMware infrastructure.
    - All variables and VMware object names are case sensitive.
    - User must hold the 'Authorization.ModifyPermissions' privilege to invoke this module.
author:
- Abhijeet Kasurde (@Akasurde)
- Christian Kotte (@ckotte)
notes:
    - Tested on ESXi 6.5
requirements:
    - "python >= 2.6"
    - PyVmomi
extends_documentation_fragment:
- community.vmware.vmware.documentation

'''

EXAMPLES = r'''
- name: Gather info about all Users on given ESXi host system
  community.vmware.vmware_local_user_info:
    hostname: '{{ esxi_hostname }}'
    username: '{{ esxi_username }}'
    password: '{{ esxi_password }}'
  delegate_to: localhost
  register: all_user_info
'''

RETURN = r'''
local_user_info:
    description: metadata about all local users
    returned: always
    type: dict
    sample: [
        {
            "role": "admin",
            "description": "Administrator",
            "group": false,
            "user_id": 0,
            "user_name": "root",
            "shell_access": true
        },
        {
            "role": "admin",
            "description": "DCUI User",
            "group": false,
            "user_id": 100,
            "user_name": "dcui",
            "shell_access": false
        },
    ]
'''

try:
    from pyVmomi import vmodl
except ImportError:
    pass

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.vmware.plugins.module_utils.vmware import PyVmomi, vmware_argument_spec
from ansible.module_utils._text import to_native


class VMwareUserInfoManager(PyVmomi):
    """Class to manage local user info"""

    def __init__(self, module):
        super(VMwareUserInfoManager, self).__init__(module)

        if self.is_vcenter():
            self.module.fail_json(
                msg="Failed to get local account manager settings.",
                details="It seems that '%s' is a vCenter server instead of an ESXi server" % self.module.params['hostname']
            )

    def gather_user_info(self):
        """Gather info about local users"""
        results = dict(changed=False, local_user_info=[])
        search_string = ''
        exact_match = False
        find_users = True
        find_groups = False
        user_accounts = self.content.userDirectory.RetrieveUserGroups(
            None, search_string, None, None, exact_match, find_users, find_groups
        )
        if user_accounts:
            for user in user_accounts:
                temp_user = dict()
                temp_user['user_name'] = user.principal
                temp_user['description'] = user.fullName
                temp_user['group'] = user.group
                temp_user['user_id'] = user.id
                temp_user['shell_access'] = user.shellAccess
                temp_user['role'] = None
                try:
                    permissions = self.content.authorizationManager.RetrieveEntityPermissions(
                        entity=self.content.rootFolder,
                        inherited=False
                    )
                except vmodl.fault.ManagedObjectNotFound as not_found:
                    self.module.fail_json(
                        msg="The entity doesn't exist" % to_native(not_found)
                    )
                for permission in permissions:
                    if permission.principal == user.principal:
                        temp_user['role'] = self.get_role_name(permission.roleId, self.content.authorizationManager.roleList)
                        break

                results['local_user_info'].append(temp_user)
        self.module.exit_json(**results)

    @staticmethod
    def get_role_name(role_id, role_list):
        """Get role name from role ID"""
        role_name = None
        # Default role: No access
        if role_id == -5:
            role_name = 'no-access'
        # Default role: Read-only
        elif role_id == -2:
            role_name = 'read-only'
        # Default role: Administrator
        elif role_id == -1:
            role_name = 'admin'
        # Custom roles
        else:
            for role in role_list:
                if role.roleId == role_id:
                    role_name = role.name
                    break
        return role_name


def main():
    """Main"""
    argument_spec = vmware_argument_spec()
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True)

    vmware_local_user_info = VMwareUserInfoManager(module)
    vmware_local_user_info.gather_user_info()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright: (c) 2018, Ansible Project
# This module is also sponsored by E.T.A.I. (www.etai.fr)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: vmware_guest_snapshot
short_description: Manages virtual machines snapshots in vCenter
description:
    - This module can be used to create, delete and update snapshot(s) of the given virtual machine.
    - All parameters and VMware object names are case sensitive.
author:
    - Loic Blot (@nerzhul) <loic.blot@unix-experience.fr>
notes:
    - Tested on vSphere 5.5, 6.0 and 6.5
requirements:
    - "python >= 2.6"
    - PyVmomi
options:
   state:
     description:
     - Manage snapshot(s) attached to a specific virtual machine.
     - If set to C(present) and snapshot absent, then will create a new snapshot with the given name.
     - If set to C(present) and snapshot present, then no changes are made.
     - If set to C(absent) and snapshot present, then snapshot with the given name is removed.
     - If set to C(absent) and snapshot absent, then no changes are made.
     - If set to C(revert) and snapshot present, then virtual machine state is reverted to the given snapshot.
     - If set to C(revert) and snapshot absent, then no changes are made.
     - If set to C(remove_all) and snapshot(s) present, then all snapshot(s) will be removed.
     - If set to C(remove_all) and snapshot(s) absent, then no changes are made.
     choices: ['present', 'absent', 'revert', 'remove_all']
     default: 'present'
     type: str
   name:
     description:
     - Name of the virtual machine to work with.
     - This is required parameter, if C(uuid) or C(moid) is not supplied.
     type: str
   name_match:
     description:
     - If multiple VMs matching the name, use the first or last found.
     default: 'first'
     choices: ['first', 'last']
     type: str
   uuid:
     description:
     - UUID of the instance to manage if known, this is VMware's BIOS UUID by default.
     - This is required if C(name) or C(moid) parameter is not supplied.
     type: str
   moid:
     description:
     - Managed Object ID of the instance to manage if known, this is a unique identifier only within a single vCenter instance.
     - This is required if C(name) or C(uuid) is not supplied.
     type: str
   use_instance_uuid:
     description:
     - Whether to use the VMware instance UUID rather than the BIOS UUID.
     default: no
     type: bool
   folder:
     description:
     - Destination folder, absolute or relative path to find an existing guest.
     - This is required parameter, if C(name) is supplied.
     - The folder should include the datacenter. ESX's datacenter is ha-datacenter.
     - 'Examples:'
     - '   folder: /ha-datacenter/vm'
     - '   folder: ha-datacenter/vm'
     - '   folder: /datacenter1/vm'
     - '   folder: datacenter1/vm'
     - '   folder: /datacenter1/vm/folder1'
     - '   folder: datacenter1/vm/folder1'
     - '   folder: /folder1/datacenter1/vm'
     - '   folder: folder1/datacenter1/vm'
     - '   folder: /folder1/datacenter1/vm/folder2'
     type: str
   datacenter:
     description:
     - Destination datacenter for the deploy operation.
     required: True
     type: str
   snapshot_name:
     description:
     - Sets the snapshot name to manage.
     - This param is required only if state is not C(remove_all)
     type: str
   description:
     description:
     - Define an arbitrary description to attach to snapshot.
     default: ''
     type: str
   quiesce:
     description:
     - If set to C(true) and virtual machine is powered on, it will quiesce the file system in virtual machine.
     - Note that VMware Tools are required for this flag.
     - If virtual machine is powered off or VMware Tools are not available, then this flag is set to C(false).
     - If virtual machine does not provide capability to take quiesce snapshot, then this flag is set to C(false).
     required: False
     type: bool
     default: False
   memory_dump:
     description:
     - If set to C(true), memory dump of virtual machine is also included in snapshot.
     - Note that memory snapshots take time and resources, this will take longer time to create.
     - If virtual machine does not provide capability to take memory snapshot, then this flag is set to C(false).
     required: False
     type: bool
     default: False
   remove_children:
     description:
     - If set to C(true) and state is set to C(absent), then entire snapshot subtree is set for removal.
     required: False
     type: bool
     default: False
   new_snapshot_name:
     description:
     - Value to rename the existing snapshot to.
     type: str
   new_description:
     description:
     - Value to change the description of an existing snapshot to.
     type: str
extends_documentation_fragment:
- community.vmware.vmware.documentation

'''

EXAMPLES = '''
  - name: Create a snapshot
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: present
      snapshot_name: snap1
      description: snap1_description
    delegate_to: localhost

  - name: Remove a snapshot
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: absent
      snapshot_name: snap1
    delegate_to: localhost

  - name: Revert to a snapshot
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: revert
      snapshot_name: snap1
    delegate_to: localhost

  - name: Remove all snapshots of a VM
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: remove_all
    delegate_to: localhost

  - name: Remove all snapshots of a VM using MoID
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      moid: vm-42
      state: remove_all
    delegate_to: localhost

  - name: Take snapshot of a VM using quiesce and memory flag on
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: present
      snapshot_name: dummy_vm_snap_0001
      quiesce: yes
      memory_dump: yes
    delegate_to: localhost

  - name: Remove a snapshot and snapshot subtree
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: absent
      remove_children: yes
      snapshot_name: snap1
    delegate_to: localhost

  - name: Rename a snapshot
    community.vmware.vmware_guest_snapshot:
      hostname: "{{ vcenter_hostname }}"
      username: "{{ vcenter_username }}"
      password: "{{ vcenter_password }}"
      datacenter: "{{ datacenter_name }}"
      folder: "/{{ datacenter_name }}/vm/"
      name: "{{ guest_name }}"
      state: present
      snapshot_name: current_snap_name
      new_snapshot_name: im_renamed
      new_description: "{{ new_snapshot_description }}"
    delegate_to: localhost
'''

RETURN = """
snapshot_results:
    description: metadata about the virtual machine snapshots
    returned: always
    type: dict
    sample: {
      "current_snapshot": {
          "creation_time": "2019-04-09T14:40:26.617427+00:00",
          "description": "Snapshot 4 example",
          "id": 4,
          "name": "snapshot4",
          "state": "poweredOff"
      },
      "snapshots": [
          {
              "creation_time": "2019-04-09T14:38:24.667543+00:00",
              "description": "Snapshot 3 example",
              "id": 3,
              "name": "snapshot3",
              "state": "poweredOff"
          },
          {
              "creation_time": "2019-04-09T14:40:26.617427+00:00",
              "description": "Snapshot 4 example",
              "id": 4,
              "name": "snapshot4",
              "state": "poweredOff"
          }
      ]
    }
"""

import time
try:
    from pyVmomi import vim
except ImportError:
    pass

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native
from ansible_collections.community.vmware.plugins.module_utils.vmware import PyVmomi, list_snapshots, vmware_argument_spec


class PyVmomiHelper(PyVmomi):
    def __init__(self, module):
        super(PyVmomiHelper, self).__init__(module)

    @staticmethod
    def wait_for_task(task):
        # https://www.vmware.com/support/developer/vc-sdk/visdk25pubs/ReferenceGuide/vim.Task.html
        # https://www.vmware.com/support/developer/vc-sdk/visdk25pubs/ReferenceGuide/vim.TaskInfo.html
        # https://github.com/virtdevninja/pyvmomi-community-samples/blob/master/samples/tools/tasks.py
        while task.info.state not in ['success', 'error']:
            time.sleep(1)

    def get_snapshots_by_name_recursively(self, snapshots, snapname):
        snap_obj = []
        for snapshot in snapshots:
            if snapshot.name == snapname:
                snap_obj.append(snapshot)
            else:
                snap_obj = snap_obj + self.get_snapshots_by_name_recursively(snapshot.childSnapshotList, snapname)
        return snap_obj

    def snapshot_vm(self, vm):
        memory_dump = False
        quiesce = False
        # Check if there is a latest snapshot already present as specified by user
        if vm.snapshot is not None:
            snap_obj = self.get_snapshots_by_name_recursively(vm.snapshot.rootSnapshotList,
                                                              self.module.params["snapshot_name"])
            if snap_obj:
                # Snapshot already exists, do not anything.
                self.module.exit_json(changed=False,
                                      msg="Snapshot named [%(snapshot_name)s] already exists and is current." % self.module.params)
        # Check if Virtual Machine provides capabilities for Quiesce and Memory Snapshots
        if vm.capability.quiescedSnapshotsSupported:
            quiesce = self.module.params['quiesce']
        if vm.capability.memorySnapshotsSupported:
            memory_dump = self.module.params['memory_dump']

        task = None
        try:
            task = vm.CreateSnapshot(self.module.params["snapshot_name"],
                                     self.module.params["description"],
                                     memory_dump,
                                     quiesce)
        except vim.fault.RestrictedVersion as exc:
            self.module.fail_json(msg="Failed to take snapshot due to VMware Licence"
                                      " restriction : %s" % to_native(exc.msg))
        except Exception as exc:
            self.module.fail_json(msg="Failed to create snapshot of virtual machine"
                                      " %s due to %s" % (self.module.params['name'], to_native(exc)))
        return task

    def rename_snapshot(self, vm):
        if vm.snapshot is None:
            vm_id = self.module.params.get('uuid') or self.module.params.get('name') or self.params.get('moid')
            self.module.fail_json(msg="virtual machine - %s doesn't have any snapshots" % vm_id)

        snap_obj = self.get_snapshots_by_name_recursively(vm.snapshot.rootSnapshotList,
                                                          self.module.params["snapshot_name"])
        task = None
        if len(snap_obj) == 1:
            snap_obj = snap_obj[0].snapshot
            if self.module.params["new_snapshot_name"] and self.module.params["new_description"]:
                task = snap_obj.RenameSnapshot(name=self.module.params["new_snapshot_name"],
                                               description=self.module.params["new_description"])
            elif self.module.params["new_snapshot_name"]:
                task = snap_obj.RenameSnapshot(name=self.module.params["new_snapshot_name"])
            else:
                task = snap_obj.RenameSnapshot(description=self.module.params["new_description"])
        else:
            vm_id = self.module.params.get('uuid') or self.module.params.get('name') or self.params.get('moid')
            self.module.exit_json(
                msg="Couldn't find any snapshots with specified name: %s on VM: %s" %
                    (self.module.params["snapshot_name"], vm_id))
        return task

    def remove_or_revert_snapshot(self, vm):
        if vm.snapshot is None:
            vm_name = (self.module.params.get('uuid') or self.module.params.get('name'))
            if self.module.params.get('state') == 'revert':
                self.module.fail_json(msg="virtual machine - %s does not"
                                          " have any snapshots to revert to." % vm_name)
            self.module.exit_json(msg="virtual machine - %s doesn't have any"
                                      " snapshots to remove." % vm_name)

        snap_obj = self.get_snapshots_by_name_recursively(vm.snapshot.rootSnapshotList,
                                                          self.module.params["snapshot_name"])
        task = None
        if len(snap_obj) == 1:
            snap_obj = snap_obj[0].snapshot
            if self.module.params["state"] == "absent":
                # Remove subtree depending upon the user input
                remove_children = self.module.params.get('remove_children', False)
                task = snap_obj.RemoveSnapshot_Task(remove_children)
            elif self.module.params["state"] == "revert":
                task = snap_obj.RevertToSnapshot_Task()
        else:
            vm_id = self.module.params.get('uuid') or self.module.params.get('name') or self.params.get('moid')
            self.module.exit_json(msg="Couldn't find any snapshots with"
                                      " specified name: %s on VM: %s" % (self.module.params["snapshot_name"], vm_id))

        return task

    def apply_snapshot_op(self, vm):
        result = {}
        if self.module.params["state"] == "present":
            if self.module.params["new_snapshot_name"] or self.module.params["new_description"]:
                self.rename_snapshot(vm)
                result = {'changed': True, 'failed': False, 'renamed': True}
                task = None
            else:
                task = self.snapshot_vm(vm)
        elif self.module.params["state"] in ["absent", "revert"]:
            task = self.remove_or_revert_snapshot(vm)
        elif self.module.params["state"] == "remove_all":
            task = vm.RemoveAllSnapshots()
        else:
            # This should not happen
            raise AssertionError()

        if task:
            self.wait_for_task(task)
            if task.info.state == 'error':
                result = {'changed': False, 'failed': True, 'msg': task.info.error.msg}
            else:
                result = {'changed': True, 'failed': False, 'snapshot_results': list_snapshots(vm)}

        return result


def main():
    argument_spec = vmware_argument_spec()
    argument_spec.update(
        state=dict(default='present', choices=['present', 'absent', 'revert', 'remove_all']),
        name=dict(type='str'),
        name_match=dict(type='str', choices=['first', 'last'], default='first'),
        uuid=dict(type='str'),
        moid=dict(type='str'),
        use_instance_uuid=dict(type='bool', default=False),
        folder=dict(type='str'),
        datacenter=dict(required=True, type='str'),
        snapshot_name=dict(type='str'),
        description=dict(type='str', default=''),
        quiesce=dict(type='bool', default=False),
        memory_dump=dict(type='bool', default=False),
        remove_children=dict(type='bool', default=False),
        new_snapshot_name=dict(type='str'),
        new_description=dict(type='str'),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        required_together=[
            ['name', 'folder']
        ],
        required_one_of=[
            ['name', 'uuid', 'moid']
        ],
    )

    if module.params['folder']:
        # FindByInventoryPath() does not require an absolute path
        # so we should leave the input folder path unmodified
        module.params['folder'] = module.params['folder'].rstrip('/')

    pyv = PyVmomiHelper(module)
    # Check if the VM exists before continuing
    vm = pyv.get_vm()

    if not vm:
        vm_id = (module.params.get('uuid') or module.params.get('name') or module.params.get('moid'))
        module.fail_json(msg="Unable to manage snapshots for non-existing VM %s" % vm_id)

    if not module.params['snapshot_name'] and module.params['state'] != 'remove_all':
        module.fail_json(msg="snapshot_name param is required when state is '%(state)s'" % module.params)

    result = pyv.apply_snapshot_op(vm)

    if 'failed' not in result:
        result['failed'] = False

    if result['failed']:
        module.fail_json(**result)
    else:
        module.exit_json(**result)


if __name__ == '__main__':
    main()



#
# (c) 2019, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#
from __future__ import absolute_import, division, print_function

__metaclass__ = type
import re

from ansible.module_utils.six import iteritems
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    to_list,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.providers.providers import (
    CliProvider,
)


class AddressFamily(CliProvider):
    def render(self, config=None):
        commands = list()
        safe_list = list()

        router_context = "router bgp %s" % self.get_value("config.bgp_as")
        context_config = None

        for item in self.get_value("config.address_family"):
            context = "address-family %s %s" % (item["afi"], item["safi"])
            context_commands = list()

            if config:
                context_path = [router_context, context]
                context_config = self.get_config_context(
                    config, context_path, indent=1
                )

            for key, value in iteritems(item):
                if value is not None:
                    meth = getattr(self, "_render_%s" % key, None)
                    if meth:
                        resp = meth(item, context_config)
                        if resp:
                            context_commands.extend(to_list(resp))

            if context_commands:
                commands.append(context)
                commands.extend(context_commands)
                commands.append("exit")

            safe_list.append(context)

        if config:
            resp = self._negate_config(config, safe_list)
            commands.extend(resp)

        return commands

    def _negate_config(self, config, safe_list=None):
        commands = list()
        matches = re.findall(r"(address-family .+)$", config, re.M)
        for item in set(matches).difference(safe_list):
            commands.append("no %s" % item)
        return commands

    def _render_networks(self, item, config=None):
        commands = list()
        safe_list = list()

        for entry in item["networks"]:
            network = entry["prefix"]
            if entry["masklen"]:
                network = "%s/%s" % (entry["prefix"], entry["masklen"])
            safe_list.append(network)

            cmd = "network %s" % network

            if entry["route_map"]:
                cmd += " route-policy %s" % entry["route_map"]

            if not config or cmd not in config:
                commands.append(cmd)

        if config and self.params["operation"] == "replace":
            matches = re.findall(r"network (\S+)", config, re.M)
            for entry in set(matches).difference(safe_list):
                commands.append("no network %s" % entry)

        return commands

    def _render_redistribute(self, item, config=None):
        commands = list()
        safe_list = list()

        for entry in item["redistribute"]:
            option = entry["protocol"]

            cmd = "redistribute %s" % entry["protocol"]

            if entry["id"] and entry["protocol"] in (
                "ospf",
                "eigrp",
                "isis",
                "ospfv3",
            ):
                cmd += " %s" % entry["id"]
                option += " %s" % entry["id"]

            if entry["metric"]:
                cmd += " metric %s" % entry["metric"]

            if entry["route_map"]:
                cmd += " route-policy %s" % entry["route_map"]

            if not config or cmd not in config:
                commands.append(cmd)

            safe_list.append(option)

        if self.params["operation"] == "replace":
            if config:
                matches = re.findall(
                    r"redistribute (\S+)(?:\s*)(\d*)", config, re.M
                )
                for i in range(0, len(matches)):
                    matches[i] = " ".join(matches[i]).strip()
                for entry in set(matches).difference(safe_list):
                    commands.append("no redistribute %s" % entry)

        return commands



#
# (c) 2018 Extreme Networks Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible_collections.community.network.tests.unit.compat.mock import patch, MagicMock
from ansible_collections.community.network.tests.unit.plugins.modules.utils import set_module_args
from ansible_collections.community.network.plugins.cliconf.exos import Cliconf
from ansible_collections.community.network.plugins.modules.network.exos import exos_config
from .exos_module import TestExosModule, load_fixture


class TestExosConfigModule(TestExosModule):

    module = exos_config

    def setUp(self):
        super(TestExosConfigModule, self).setUp()

        self.mock_get_config = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.get_config')
        self.get_config = self.mock_get_config.start()

        self.mock_load_config = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.load_config')
        self.load_config = self.mock_load_config.start()

        self.mock_run_commands = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.run_commands')
        self.run_commands = self.mock_run_commands.start()

        self.mock_get_startup_config = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.get_startup_config')
        self.get_startup_config = self.mock_get_startup_config.start()

        self.cliconf_obj = Cliconf(MagicMock())

        self.mock_get_diff = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.get_diff')
        self.get_diff = self.mock_get_diff.start()

        self.running_config = load_fixture('exos_config_config.cfg')

    def tearDown(self):
        super(TestExosConfigModule, self).tearDown()
        self.mock_get_config.stop()
        self.mock_load_config.stop()
        self.mock_run_commands.stop()
        self.mock_get_startup_config.stop()

    def load_fixtures(self, commands=None):
        config_file = 'exos_config_config.cfg'
        self.get_config.return_value = load_fixture(config_file)
        self.load_config.return_value = None

    def test_exos_config_unchanged(self):
        src = load_fixture('exos_config_config.cfg')
        set_module_args(dict(src=src))
        self.get_diff.return_value = self.cliconf_obj.get_diff(src, src)
        self.execute_module()

    def test_exos_config_src(self):
        src = load_fixture('exos_config_src.cfg')
        set_module_args(dict(src=src))
        commands = ['configure ports 1 description-string "IDS"',
                    'configure snmp sysName "marble"']
        self.get_diff.return_value = self.cliconf_obj.get_diff(src, self.running_config)
        self.execute_module(changed=True, commands=commands)

    def test_exos_config_backup(self):
        set_module_args(dict(backup=True))
        result = self.execute_module()
        self.assertIn('__backup__', result)

    def test_exos_config_save_always(self):
        self.run_commands.return_value = 'configure snmp sysName "marble"'
        set_module_args(dict(save_when='always'))
        self.execute_module(changed=True)
        self.assertEqual(self.run_commands.call_count, 1)
        self.assertEqual(self.get_config.call_count, 0)
        self.assertEqual(self.load_config.call_count, 0)
        args = self.run_commands.call_args[0][1]
        self.assertIn('save configuration', args['command'])

    def test_exos_config_save_changed_true(self):
        src = load_fixture('exos_config_src.cfg')
        set_module_args(dict(src=src, save_when='changed'))
        commands = ['configure ports 1 description-string "IDS"',
                    'configure snmp sysName "marble"']
        self.get_diff.return_value = self.cliconf_obj.get_diff(src, self.running_config)
        self.execute_module(changed=True, commands=commands)
        self.assertEqual(self.run_commands.call_count, 1)
        self.assertEqual(self.get_config.call_count, 1)
        self.assertEqual(self.load_config.call_count, 1)
        args = self.run_commands.call_args[0][1]
        self.assertIn('save configuration', args['command'])

    def test_exos_config_save_changed_true_check_mode(self):
        src = load_fixture('exos_config_src.cfg')
        set_module_args(dict(src=src, save_when='changed', _ansible_check_mode=True))
        commands = ['configure ports 1 description-string "IDS"',
                    'configure snmp sysName "marble"']
        self.get_diff.return_value = self.cliconf_obj.get_diff(src, self.running_config)
        self.execute_module(changed=True, commands=commands)
        self.assertEqual(self.run_commands.call_count, 0)
        self.assertEqual(self.get_config.call_count, 1)
        self.assertEqual(self.load_config.call_count, 0)

    def test_exos_config_save_changed_false(self):
        set_module_args(dict(save_when='changed'))
        self.execute_module(changed=False)
        self.assertEqual(self.run_commands.call_count, 0)
        self.assertEqual(self.get_config.call_count, 0)
        self.assertEqual(self.load_config.call_count, 0)

    def test_exos_config_save_modified_false(self):
        self.get_startup_config.return_value = load_fixture('exos_config_config.cfg')
        set_module_args(dict(save_when='modified'))
        self.execute_module(changed=False)
        self.assertEqual(self.run_commands.call_count, 0)
        self.assertEqual(self.get_config.call_count, 1)
        self.assertEqual(self.get_startup_config.call_count, 1)
        self.assertEqual(self.load_config.call_count, 0)

    def test_exos_config_save_modified_true(self):
        self.get_startup_config.return_value = load_fixture('exos_config_modified.cfg')
        set_module_args(dict(save_when='modified'))
        self.execute_module(changed=True)
        self.assertEqual(self.run_commands.call_count, 1)
        self.assertTrue(self.get_config.call_count > 0)
        self.assertEqual(self.get_startup_config.call_count, 1)
        self.assertEqual(self.load_config.call_count, 0)

    def test_exos_config_lines(self):
        lines = ['configure snmp sysName "marble"']
        set_module_args(dict(lines=lines))
        commands = ['configure snmp sysName "marble"']
        self.get_diff.return_value = self.cliconf_obj.get_diff('\n'.join(lines), self.running_config)
        self.execute_module(changed=True, commands=commands)

    def test_exos_config_before(self):
        lines = ['configure snmp sysName "marble"']
        set_module_args(dict(lines=lines, before=['test1', 'test2']))
        commands = ['test1', 'test2', 'configure snmp sysName "marble"']
        self.get_diff.return_value = self.cliconf_obj.get_diff('\n'.join(lines), self.running_config)
        self.execute_module(changed=True, commands=commands, sort=False)

    def test_exos_config_after(self):
        lines = ['configure snmp sysName "marble"']
        set_module_args(dict(lines=lines, after=['test1', 'test2']))
        commands = ['configure snmp sysName "marble"', 'test1', 'test2']
        self.get_diff.return_value = self.cliconf_obj.get_diff('\n'.join(lines), self.running_config)
        self.execute_module(changed=True, commands=commands, sort=False)

    def test_exos_config_before_after_no_change(self):
        lines = ['configure snmp sysName "x870"']
        set_module_args(dict(lines=lines,
                             before=['test1', 'test2'],
                             after=['test3', 'test4']))
        self.get_diff.return_value = self.cliconf_obj.get_diff('\n'.join(lines), self.running_config)
        self.execute_module()

    def test_exos_config_config(self):
        config = 'hostname localhost'
        lines = ['configure snmp sysName "x870"']
        set_module_args(dict(lines=lines, config=config))
        commands = ['configure snmp sysName "x870"']
        self.get_diff.return_value = self.cliconf_obj.get_diff('\n'.join(lines), config)
        self.execute_module(changed=True, commands=commands)

    def test_exos_config_match_none(self):
        lines = ['configure snmp sysName "x870"']
        set_module_args(dict(lines=lines, match='none'))
        self.get_diff.return_value = self.cliconf_obj.get_diff('\n'.join(lines), self.running_config, diff_match='none')
        self.execute_module(changed=True, commands=lines)

    def test_exos_config_src_and_lines_fails(self):
        args = dict(src='foo', lines='foo')
        set_module_args(args)
        self.execute_module(failed=True)

    def test_exos_config_match_exact_requires_lines(self):
        args = dict(match='exact')
        set_module_args(args)
        self.execute_module(failed=True)

    def test_exos_config_match_strict_requires_lines(self):
        args = dict(match='strict')
        set_module_args(args)
        self.execute_module(failed=True)

    def test_exos_config_replace_block_requires_lines(self):
        args = dict(replace='block')
        set_module_args(args)
        self.execute_module(failed=True)

    def test_exos_config_replace_config_requires_src(self):
        args = dict(replace='config')
        set_module_args(args)
        self.execute_module(failed=True)

    def test_exos_diff_running_unchanged(self):
        args = dict(diff_against='running', _ansible_diff=True)
        set_module_args(args)
        self.execute_module(changed=False)

    def test_exos_diff_running_unchanged_check(self):
        args = dict(diff_against='running',
                    _ansible_diff=True,
                    _ansible_check_mode=True)
        set_module_args(args)
        self.execute_module(changed=False)

    def test_exos_diff_startup_unchanged(self):
        mock_get_startup_config = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.get_startup_config')
        get_startup_config = mock_get_startup_config.start()
        get_startup_config.return_value = load_fixture('exos_config_config.cfg')

        args = dict(diff_against='startup', _ansible_diff=True)
        set_module_args(args)
        self.execute_module(changed=False)
        self.assertEqual(get_startup_config.call_count, 1)

        mock_get_startup_config.stop()

    def test_exos_diff_startup_changed(self):
        mock_get_startup_config = patch('ansible_collections.community.network.plugins.modules.network.exos.exos_config.get_startup_config')
        get_startup_config = mock_get_startup_config.start()
        get_startup_config.return_value = load_fixture('exos_config_modified.cfg')

        args = dict(diff_against='startup', _ansible_diff=True)
        set_module_args(args)
        self.execute_module(changed=True)
        self.assertEqual(get_startup_config.call_count, 1)

        mock_get_startup_config.stop()

    def test_exos_diff_intended_unchanged(self):
        intended_config = load_fixture('exos_config_config.cfg')
        args = dict(diff_against='intended',
                    intended_config=intended_config,
                    _ansible_diff=True)
        set_module_args(args)
        self.get_diff = self.cliconf_obj.get_diff(intended_config, self.running_config)
        self.execute_module(changed=False)

    def test_exos_diff_intended_modified(self):
        intended_config = load_fixture('exos_config_modified.cfg')
        args = dict(diff_against='intended',
                    intended_config=intended_config,
                    _ansible_diff=True)
        set_module_args(args)
        self.get_diff = self.cliconf_obj.get_diff(intended_config, self.running_config)
        self.execute_module(changed=True)



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
---
module: ce_sflow
short_description: Manages sFlow configuration on HUAWEI CloudEngine switches.
description:
    - Configure Sampled Flow (sFlow) to monitor traffic on an interface in real time,
      detect abnormal traffic, and locate the source of attack traffic,
      ensuring stable running of the network.
author: QijunPan (@QijunPan)
notes:
    - This module requires the netconf system service be enabled on the remote device being managed.
    - Recommended connection is C(netconf).
    - This module also works with C(local) connections for legacy playbooks.
options:
    agent_ip:
        description:
            - Specifies the IPv4/IPv6 address of an sFlow agent.
    source_ip:
        description:
            - Specifies the source IPv4/IPv6 address of sFlow packets.
    collector_id:
        description:
            - Specifies the ID of an sFlow collector. This ID is used when you specify
              the collector in subsequent sFlow configuration.
        choices: ['1', '2']
    collector_ip:
        description:
            - Specifies the IPv4/IPv6 address of the sFlow collector.
    collector_ip_vpn:
        description:
            - Specifies the name of a VPN instance.
              The value is a string of 1 to 31 case-sensitive characters, spaces not supported.
              When double quotation marks are used around the string, spaces are allowed in the string.
              The value C(_public_) is reserved and cannot be used as the VPN instance name.
    collector_datagram_size:
        description:
            - Specifies the maximum length of sFlow packets sent from an sFlow agent to an sFlow collector.
              The value is an integer, in bytes. It ranges from 1024 to 8100. The default value is 1400.
    collector_udp_port:
        description:
            - Specifies the UDP destination port number of sFlow packets.
              The value is an integer that ranges from 1 to 65535. The default value is 6343.
    collector_meth:
        description:
            - Configures the device to send sFlow packets through service interfaces,
              enhancing the sFlow packet forwarding capability.
              The enhanced parameter is optional. No matter whether you configure the enhanced mode,
              the switch determines to send sFlow packets through service cards or management port
              based on the routing information on the collector.
              When the value is meth, the device forwards sFlow packets at the control plane.
              When the value is enhanced, the device forwards sFlow packets at the forwarding plane to
              enhance the sFlow packet forwarding capacity.
        choices: ['meth', 'enhanced']
    collector_description:
        description:
            - Specifies the description of an sFlow collector.
              The value is a string of 1 to 255 case-sensitive characters without spaces.
    sflow_interface:
        description:
            - Full name of interface for Flow Sampling or Counter.
              It must be a physical interface, Eth-Trunk, or Layer 2 subinterface.
    sample_collector:
        description:
            -  Indicates the ID list of the collector.
    sample_rate:
        description:
            - Specifies the flow sampling rate in the format 1/rate.
              The value is an integer and ranges from 1 to 4294967295. The default value is 8192.
    sample_length:
        description:
            - Specifies the maximum length of sampled packets.
              The value is an integer and ranges from 18 to 512, in bytes. The default value is 128.
    sample_direction:
        description:
            - Enables flow sampling in the inbound or outbound direction.
        choices: ['inbound', 'outbound', 'both']
    counter_collector:
        description:
            - Indicates the ID list of the counter collector.
    counter_interval:
        description:
            - Indicates the counter sampling interval.
              The value is an integer that ranges from 10 to 4294967295, in seconds. The default value is 20.
    export_route:
        description:
            - Configures the sFlow packets sent by the switch not to carry routing information.
        choices: ['enable', 'disable']
    rate_limit:
        description:
            - Specifies the rate of sFlow packets sent from a card to the control plane.
              The value is an integer that ranges from 100 to 1500, in pps.
        type: str
        version_added: '0.2.0'
    rate_limit_slot:
        description:
            - Specifies the slot where the rate of output sFlow packets is limited.
              If this parameter is not specified, the rate of sFlow packets sent from
              all cards to the control plane is limited.
              The value is an integer or a string of characters.
        type: str
        version_added: '0.2.0'
    forward_enp_slot:
        description:
            - Enable the Embedded Network Processor (ENP) chip function.
              The switch uses the ENP chip to perform sFlow sampling,
              and the maximum sFlow sampling interval is 65535.
              If you set the sampling interval to be larger than 65535,
              the switch automatically restores it to 65535.
              The value is an integer or 'all'.
        type: str
        version_added: '0.2.0'
    state:
        description:
            - Determines whether the config should be present or not
              on the device.
        default: present
        choices: ['present', 'absent']
'''

EXAMPLES = '''
---

- name: Sflow module test
  hosts: ce128
  connection: local
  gather_facts: no
  vars:
    cli:
      host: "{{ inventory_hostname }}"
      port: "{{ ansible_ssh_port }}"
      username: "{{ username }}"
      password: "{{ password }}"
      transport: cli

  tasks:
  - name: Configuring sFlow Agent
    community.network.ce_sflow:
      agent_ip: 6.6.6.6
      provider: '{{ cli }}'

  - name: Configuring sFlow Collector
    community.network.ce_sflow:
      collector_id: 1
      collector_ip: 7.7.7.7
      collector_ip_vpn: vpn1
      collector_description: Collector1
      provider: '{{ cli }}'

  - name: Configure flow sampling.
    community.network.ce_sflow:
      sflow_interface: 10GE2/0/2
      sample_collector: 1
      sample_direction: inbound
      provider: '{{ cli }}'

  - name: Configure counter sampling.
    community.network.ce_sflow:
      sflow_interface: 10GE2/0/2
      counter_collector: 1
      counter_interval: 1000
      provider: '{{ cli }}'
'''

RETURN = '''
proposed:
    description: k/v pairs of parameters passed into module
    returned: verbose mode
    type: dict
    sample: {"agent_ip": "6.6.6.6", "state": "present"}
existing:
    description: k/v pairs of existing configuration
    returned: verbose mode
    type: dict
    sample: {"agent": {}}
end_state:
    description: k/v pairs of configuration after module execution
    returned: verbose mode
    type: dict
    sample: {"agent": {"family": "ipv4", "ipv4Addr": "1.2.3.4", "ipv6Addr": null}}
updates:
    description: commands sent to the device
    returned: always
    type: list
    sample: ["sflow agent ip 6.6.6.6"]
changed:
    description: check to see if a change was made on the device
    returned: always
    type: bool
    sample: true
'''

import re
from xml.etree import ElementTree
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.network.plugins.module_utils.network.cloudengine.ce import get_nc_config, set_nc_config, ce_argument_spec, check_ip_addr

CE_NC_GET_SFLOW = """
<filter type="subtree">
<sflow xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <sources>
        <source>
            <family></family>
            <ipv4Addr></ipv4Addr>
            <ipv6Addr></ipv6Addr>
        </source>
    </sources>
    <agents>
        <agent>
            <family></family>
            <ipv4Addr></ipv4Addr>
            <ipv6Addr></ipv6Addr>
        </agent>
    </agents>
    <collectors>
        <collector>
            <collectorID></collectorID>
            <family></family>
            <ipv4Addr></ipv4Addr>
            <ipv6Addr></ipv6Addr>
            <vrfName></vrfName>
            <datagramSize></datagramSize>
            <port></port>
            <description></description>
            <meth></meth>
        </collector>
    </collectors>
    <samplings>
        <sampling>
            <ifName>%s</ifName>
            <collectorID></collectorID>
            <direction></direction>
            <length></length>
            <rate></rate>
        </sampling>
    </samplings>
    <counters>
        <counter>
            <ifName>%s</ifName>
            <collectorID></collectorID>
            <interval></interval>
        </counter>
    </counters>
    <exports>
        <export>
            <ExportRoute></ExportRoute>
        </export>
    </exports>
</sflow>
</filter>
"""


def is_config_exist(cmp_cfg, test_cfg):
    """is configuration exist?"""

    if not cmp_cfg or not test_cfg:
        return False

    return bool(test_cfg in cmp_cfg)


def is_valid_ip_vpn(vpname):
    """check ip vpn"""

    if not vpname:
        return False

    if vpname == "_public_":
        return False

    if len(vpname) < 1 or len(vpname) > 31:
        return False

    return True


def get_ip_version(address):
    """get ip version fast"""

    if not address:
        return None

    if address.count(':') >= 2 and address.count(":") <= 7:
        return "ipv6"
    elif address.count('.') == 3:
        return "ipv4"
    else:
        return None


def get_interface_type(interface):
    """get the type of interface, such as 10GE, ETH-TRUNK, VLANIF..."""

    if interface is None:
        return None

    if interface.upper().startswith('GE'):
        iftype = 'ge'
    elif interface.upper().startswith('10GE'):
        iftype = '10ge'
    elif interface.upper().startswith('25GE'):
        iftype = '25ge'
    elif interface.upper().startswith('4X10GE'):
        iftype = '4x10ge'
    elif interface.upper().startswith('40GE'):
        iftype = '40ge'
    elif interface.upper().startswith('100GE'):
        iftype = '100ge'
    elif interface.upper().startswith('VLANIF'):
        iftype = 'vlanif'
    elif interface.upper().startswith('LOOPBACK'):
        iftype = 'loopback'
    elif interface.upper().startswith('METH'):
        iftype = 'meth'
    elif interface.upper().startswith('ETH-TRUNK'):
        iftype = 'eth-trunk'
    elif interface.upper().startswith('VBDIF'):
        iftype = 'vbdif'
    elif interface.upper().startswith('NVE'):
        iftype = 'nve'
    elif interface.upper().startswith('TUNNEL'):
        iftype = 'tunnel'
    elif interface.upper().startswith('ETHERNET'):
        iftype = 'ethernet'
    elif interface.upper().startswith('FCOE-PORT'):
        iftype = 'fcoe-port'
    elif interface.upper().startswith('FABRIC-PORT'):
        iftype = 'fabric-port'
    elif interface.upper().startswith('STACK-PORT'):
        iftype = 'stack-port'
    elif interface.upper().startswith('NULL'):
        iftype = 'null'
    else:
        return None

    return iftype.lower()


class Sflow(object):
    """Manages sFlow"""

    def __init__(self, argument_spec):
        self.spec = argument_spec
        self.module = None
        self.__init_module__()

        # module input info
        self.agent_ip = self.module.params['agent_ip']
        self.agent_version = None
        self.source_ip = self.module.params['source_ip']
        self.source_version = None
        self.export_route = self.module.params['export_route']
        self.rate_limit = self.module.params['rate_limit']
        self.rate_limit_slot = self.module.params['rate_limit_slot']
        self.forward_enp_slot = self.module.params['forward_enp_slot']
        self.collector_id = self.module.params['collector_id']
        self.collector_ip = self.module.params['collector_ip']
        self.collector_version = None
        self.collector_ip_vpn = self.module.params['collector_ip_vpn']
        self.collector_datagram_size = self.module.params['collector_datagram_size']
        self.collector_udp_port = self.module.params['collector_udp_port']
        self.collector_meth = self.module.params['collector_meth']
        self.collector_description = self.module.params['collector_description']
        self.sflow_interface = self.module.params['sflow_interface']
        self.sample_collector = self.module.params['sample_collector'] or list()
        self.sample_rate = self.module.params['sample_rate']
        self.sample_length = self.module.params['sample_length']
        self.sample_direction = self.module.params['sample_direction']
        self.counter_collector = self.module.params['counter_collector'] or list()
        self.counter_interval = self.module.params['counter_interval']
        self.state = self.module.params['state']

        # state
        self.config = ""  # current config
        self.sflow_dict = dict()
        self.changed = False
        self.updates_cmd = list()
        self.commands = list()
        self.results = dict()
        self.proposed = dict()
        self.existing = dict()
        self.end_state = dict()

    def __init_module__(self):
        """init module"""

        required_together = [("collector_id", "collector_ip")]
        self.module = AnsibleModule(
            argument_spec=self.spec, required_together=required_together, supports_check_mode=True)

    def check_response(self, con_obj, xml_name):
        """Check if response message is already succeed"""

        xml_str = con_obj.xml
        if "<ok/>" not in xml_str:
            self.module.fail_json(msg='Error: %s failed.' % xml_name)

    def netconf_set_config(self, xml_str, xml_name):
        """netconf set config"""

        rcv_xml = set_nc_config(self.module, xml_str)
        if "<ok/>" not in rcv_xml:
            self.module.fail_json(msg='Error: %s failed.' % xml_name)

    def get_sflow_dict(self):
        """ sflow config dict"""

        sflow_dict = dict(source=list(), agent=dict(), collector=list(),
                          sampling=dict(), counter=dict(), export=dict())
        conf_str = CE_NC_GET_SFLOW % (
            self.sflow_interface, self.sflow_interface)

        if not self.collector_meth:
            conf_str = conf_str.replace("<meth></meth>", "")

        rcv_xml = get_nc_config(self.module, conf_str)

        if "<data/>" in rcv_xml:
            return sflow_dict

        xml_str = rcv_xml.replace('\r', '').replace('\n', '').\
            replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', "").\
            replace('xmlns="http://www.huawei.com/netconf/vrp"', "")
        root = ElementTree.fromstring(xml_str)

        # get source info
        srcs = root.findall("sflow/sources/source")
        if srcs:
            for src in srcs:
                attrs = dict()
                for attr in src:
                    if attr.tag in ["family", "ipv4Addr", "ipv6Addr"]:
                        attrs[attr.tag] = attr.text
                sflow_dict["source"].append(attrs)

        # get agent info
        agent = root.find("sflow/agents/agent")
        if agent:
            for attr in agent:
                if attr.tag in ["family", "ipv4Addr", "ipv6Addr"]:
                    sflow_dict["agent"][attr.tag] = attr.text

        # get collector info
        collectors = root.findall("sflow/collectors/collector")
        if collectors:
            for collector in collectors:
                attrs = dict()
                for attr in collector:
                    if attr.tag in ["collectorID", "family", "ipv4Addr", "ipv6Addr",
                                    "vrfName", "datagramSize", "port", "description", "meth"]:
                        attrs[attr.tag] = attr.text
                sflow_dict["collector"].append(attrs)

        # get sampling info
        sample = root.find("sflow/samplings/sampling")
        if sample:
            for attr in sample:
                if attr.tag in ["ifName", "collectorID", "direction", "length", "rate"]:
                    sflow_dict["sampling"][attr.tag] = attr.text

        # get counter info
        counter = root.find("sflow/counters/counter")
        if counter:
            for attr in counter:
                if attr.tag in ["ifName", "collectorID", "interval"]:
                    sflow_dict["counter"][attr.tag] = attr.text

        # get export info
        export = root.find("sflow/exports/export")
        if export:
            for attr in export:
                if attr.tag == "ExportRoute":
                    sflow_dict["export"][attr.tag] = attr.text

        return sflow_dict

    def config_agent(self):
        """configures sFlow agent"""

        xml_str = ''
        if not self.agent_ip:
            return xml_str

        self.agent_version = get_ip_version(self.agent_ip)
        if not self.agent_version:
            self.module.fail_json(msg="Error: agent_ip is invalid.")

        if self.state == "present":
            if self.agent_ip != self.sflow_dict["agent"].get("ipv4Addr") \
                    and self.agent_ip != self.sflow_dict["agent"].get("ipv6Addr"):
                xml_str += '<agents><agent operation="merge">'
                xml_str += '<family>%s</family>' % self.agent_version
                if self.agent_version == "ipv4":
                    xml_str += '<ipv4Addr>%s</ipv4Addr>' % self.agent_ip
                    self.updates_cmd.append("sflow agent ip %s" % self.agent_ip)
                else:
                    xml_str += '<ipv6Addr>%s</ipv6Addr>' % self.agent_ip
                    self.updates_cmd.append("sflow agent ipv6 %s" % self.agent_ip)
                xml_str += '</agent></agents>'

        else:
            if self.agent_ip == self.sflow_dict["agent"].get("ipv4Addr") \
                    or self.agent_ip == self.sflow_dict["agent"].get("ipv6Addr"):
                xml_str += '<agents><agent operation="delete"></agent></agents>'
                self.updates_cmd.append("undo sflow agent")

        return xml_str

    def config_source(self):
        """configures the source IP address for sFlow packets"""

        xml_str = ''
        if not self.source_ip:
            return xml_str

        self.source_version = get_ip_version(self.source_ip)
        if not self.source_version:
            self.module.fail_json(msg="Error: source_ip is invalid.")

        src_dict = dict()
        for src in self.sflow_dict["source"]:
            if src.get("family") == self.source_version:
                src_dict = src
                break

        if self.state == "present":
            if self.source_ip != src_dict.get("ipv4Addr") \
                    and self.source_ip != src_dict.get("ipv6Addr"):
                xml_str += '<sources><source operation="merge">'
                xml_str += '<family>%s</family>' % self.source_version
                if self.source_version == "ipv4":
                    xml_str += '<ipv4Addr>%s</ipv4Addr>' % self.source_ip
                    self.updates_cmd.append("sflow source ip %s" % self.source_ip)
                else:
                    xml_str += '<ipv6Addr>%s</ipv6Addr>' % self.source_ip
                    self.updates_cmd.append(
                        "sflow source ipv6 %s" % self.source_ip)
                xml_str += '</source ></sources>'
        else:
            if self.source_ip == src_dict.get("ipv4Addr"):
                xml_str += '<sources><source operation="delete"><family>ipv4</family></source ></sources>'
                self.updates_cmd.append("undo sflow source ip %s" % self.source_ip)
            elif self.source_ip == src_dict.get("ipv6Addr"):
                xml_str += '<sources><source operation="delete"><family>ipv6</family></source ></sources>'
                self.updates_cmd.append("undo sflow source ipv6 %s" % self.source_ip)

        return xml_str

    def config_collector(self):
        """creates an sFlow collector and sets or modifies optional parameters for the sFlow collector"""

        xml_str = ''
        if not self.collector_id:
            return xml_str

        if self.state == "present" and not self.collector_ip:
            return xml_str

        if self.collector_ip:
            self.collector_version = get_ip_version(self.collector_ip)
            if not self.collector_version:
                self.module.fail_json(msg="Error: collector_ip is invalid.")

        # get collector dict
        exist_dict = dict()
        for collector in self.sflow_dict["collector"]:
            if collector.get("collectorID") == self.collector_id:
                exist_dict = collector
                break

        change = False
        if self.state == "present":
            if not exist_dict:
                change = True
            elif self.collector_version != exist_dict.get("family"):
                change = True
            elif self.collector_version == "ipv4" and self.collector_ip != exist_dict.get("ipv4Addr"):
                change = True
            elif self.collector_version == "ipv6" and self.collector_ip != exist_dict.get("ipv6Addr"):
                change = True
            elif self.collector_ip_vpn and self.collector_ip_vpn != exist_dict.get("vrfName"):
                change = True
            elif not self.collector_ip_vpn and exist_dict.get("vrfName") != "_public_":
                change = True
            elif self.collector_udp_port and self.collector_udp_port != exist_dict.get("port"):
                change = True
            elif not self.collector_udp_port and exist_dict.get("port") != "6343":
                change = True
            elif self.collector_datagram_size and self.collector_datagram_size != exist_dict.get("datagramSize"):
                change = True
            elif not self.collector_datagram_size and exist_dict.get("datagramSize") != "1400":
                change = True
            elif self.collector_meth and self.collector_meth != exist_dict.get("meth"):
                change = True
            elif not self.collector_meth and exist_dict.get("meth") and exist_dict.get("meth") != "meth":
                change = True
            elif self.collector_description and self.collector_description != exist_dict.get("description"):
                change = True
            elif not self.collector_description and exist_dict.get("description"):
                change = True
            else:
                pass
        else:  # absent
            # collector not exist
            if not exist_dict:
                return xml_str
            if self.collector_version and self.collector_version != exist_dict.get("family"):
                return xml_str
            if self.collector_version == "ipv4" and self.collector_ip != exist_dict.get("ipv4Addr"):
                return xml_str
            if self.collector_version == "ipv6" and self.collector_ip != exist_dict.get("ipv6Addr"):
                return xml_str
            if self.collector_ip_vpn and self.collector_ip_vpn != exist_dict.get("vrfName"):
                return xml_str
            if self.collector_udp_port and self.collector_udp_port != exist_dict.get("port"):
                return xml_str
            if self.collector_datagram_size and self.collector_datagram_size != exist_dict.get("datagramSize"):
                return xml_str
            if self.collector_meth and self.collector_meth != exist_dict.get("meth"):
                return xml_str
            if self.collector_description and self.collector_description != exist_dict.get("description"):
                return xml_str
            change = True

        if not change:
            return xml_str

        # update or delete
        if self.state == "absent":
            xml_str += '<collectors><collector operation="delete"><collectorID>%s</collectorID>' % self.collector_id
            self.updates_cmd.append("undo collector %s" % self.collector_id)
        else:
            xml_str += '<collectors><collector operation="merge"><collectorID>%s</collectorID>' % self.collector_id
            cmd = "sflow collector %s" % self.collector_id
            xml_str += '<family>%s</family>' % self.collector_version
            if self.collector_version == "ipv4":
                cmd += " ip %s" % self.collector_ip
                xml_str += '<ipv4Addr>%s</ipv4Addr>' % self.collector_ip
            else:
                cmd += " ipv6 %s" % self.collector_ip
                xml_str += '<ipv6Addr>%s</ipv6Addr>' % self.collector_ip
            if self.collector_ip_vpn:
                cmd += " vpn-instance %s" % self.collector_ip_vpn
                xml_str += '<vrfName>%s</vrfName>' % self.collector_ip_vpn
            if self.collector_datagram_size:
                cmd += " length %s" % self.collector_datagram_size
                xml_str += '<datagramSize>%s</datagramSize>' % self.collector_datagram_size
            if self.collector_udp_port:
                cmd += " udp-port %s" % self.collector_udp_port
                xml_str += '<port>%s</port>' % self.collector_udp_port
            if self.collector_description:
                cmd += " description %s" % self.collector_description
                xml_str += '<description>%s</description>' % self.collector_description
            else:
                xml_str += '<description></description>'
            if self.collector_meth:
                if self.collector_meth == "enhanced":
                    cmd += " enhanced"
                xml_str += '<meth>%s</meth>' % self.collector_meth
            self.updates_cmd.append(cmd)

        xml_str += "</collector></collectors>"

        return xml_str

    def config_sampling(self):
        """configure sflow sampling on an interface"""

        xml_str = ''
        if not self.sflow_interface:
            return xml_str

        if not self.sflow_dict["sampling"] and self.state == "absent":
            return xml_str

        self.updates_cmd.append("interface %s" % self.sflow_interface)
        if self.state == "present":
            xml_str += '<samplings><sampling operation="merge"><ifName>%s</ifName>' % self.sflow_interface
        else:
            xml_str += '<samplings><sampling operation="delete"><ifName>%s</ifName>' % self.sflow_interface

        # sample_collector
        if self.sample_collector:
            if self.sflow_dict["sampling"].get("collectorID") \
                    and self.sflow_dict["sampling"].get("collectorID") != "invalid":
                existing = self.sflow_dict["sampling"].get("collectorID").split(',')
            else:
                existing = list()

            if self.state == "present":
                diff = list(set(self.sample_collector) - set(existing))
                if diff:
                    self.updates_cmd.append(
                        "sflow sampling collector %s" % ' '.join(diff))
                    new_set = list(self.sample_collector + existing)
                    xml_str += '<collectorID>%s</collectorID>' % ','.join(list(set(new_set)))
            else:
                same = list(set(self.sample_collector) & set(existing))
                if same:
                    self.updates_cmd.append(
                        "undo sflow sampling collector %s" % ' '.join(same))
                    xml_str += '<collectorID>%s</collectorID>' % ','.join(list(set(same)))

        # sample_rate
        if self.sample_rate:
            exist = bool(self.sample_rate == self.sflow_dict["sampling"].get("rate"))
            if self.state == "present" and not exist:
                self.updates_cmd.append(
                    "sflow sampling rate %s" % self.sample_rate)
                xml_str += '<rate>%s</rate>' % self.sample_rate
            elif self.state == "absent" and exist:
                self.updates_cmd.append(
                    "undo sflow sampling rate %s" % self.sample_rate)
                xml_str += '<rate>%s</rate>' % self.sample_rate

        # sample_length
        if self.sample_length:
            exist = bool(self.sample_length == self.sflow_dict["sampling"].get("length"))
            if self.state == "present" and not exist:
                self.updates_cmd.append(
                    "sflow sampling length %s" % self.sample_length)
                xml_str += '<length>%s</length>' % self.sample_length
            elif self.state == "absent" and exist:
                self.updates_cmd.append(
                    "undo sflow sampling length %s" % self.sample_length)
                xml_str += '<length>%s</length>' % self.sample_length

        # sample_direction
        if self.sample_direction:
            direction = list()
            if self.sample_direction == "both":
                direction = ["inbound", "outbound"]
            else:
                direction.append(self.sample_direction)
            existing = list()
            if self.sflow_dict["sampling"].get("direction"):
                if self.sflow_dict["sampling"].get("direction") == "both":
                    existing = ["inbound", "outbound"]
                else:
                    existing.append(
                        self.sflow_dict["sampling"].get("direction"))

            if self.state == "present":
                diff = list(set(direction) - set(existing))
                if diff:
                    new_set = list(set(direction + existing))
                    self.updates_cmd.append(
                        "sflow sampling %s" % ' '.join(diff))
                    if len(new_set) > 1:
                        new_dir = "both"
                    else:
                        new_dir = new_set[0]
                    xml_str += '<direction>%s</direction>' % new_dir
            else:
                same = list(set(existing) & set(direction))
                if same:
                    self.updates_cmd.append("undo sflow sampling %s" % ' '.join(same))
                    if len(same) > 1:
                        del_dir = "both"
                    else:
                        del_dir = same[0]
                    xml_str += '<direction>%s</direction>' % del_dir

        if xml_str.endswith("</ifName>"):
            self.updates_cmd.pop()
            return ""

        xml_str += '</sampling></samplings>'

        return xml_str

    def config_counter(self):
        """configures sflow counter on an interface"""

        xml_str = ''
        if not self.sflow_interface:
            return xml_str

        if not self.sflow_dict["counter"] and self.state == "absent":
            return xml_str

        self.updates_cmd.append("interface %s" % self.sflow_interface)
        if self.state == "present":
            xml_str += '<counters><counter operation="merge"><ifName>%s</ifName>' % self.sflow_interface
        else:
            xml_str += '<counters><counter operation="delete"><ifName>%s</ifName>' % self.sflow_interface

        # counter_collector
        if self.counter_collector:
            if self.sflow_dict["counter"].get("collectorID") \
                    and self.sflow_dict["counter"].get("collectorID") != "invalid":
                existing = self.sflow_dict["counter"].get("collectorID").split(',')
            else:
                existing = list()

            if self.state == "present":
                diff = list(set(self.counter_collector) - set(existing))
                if diff:
                    self.updates_cmd.append("sflow counter collector %s" % ' '.join(diff))
                    new_set = list(self.counter_collector + existing)
                    xml_str += '<collectorID>%s</collectorID>' % ','.join(list(set(new_set)))
            else:
                same = list(set(self.counter_collector) & set(existing))
                if same:
                    self.updates_cmd.append(
                        "undo sflow counter collector %s" % ' '.join(same))
                    xml_str += '<collectorID>%s</collectorID>' % ','.join(list(set(same)))

        # counter_interval
        if self.counter_interval:
            exist = bool(self.counter_interval == self.sflow_dict["counter"].get("interval"))
            if self.state == "present" and not exist:
                self.updates_cmd.append(
                    "sflow counter interval %s" % self.counter_interval)
                xml_str += '<interval>%s</interval>' % self.counter_interval
            elif self.state == "absent" and exist:
                self.updates_cmd.append(
                    "undo sflow counter interval %s" % self.counter_interval)
                xml_str += '<interval>%s</interval>' % self.counter_interval

        if xml_str.endswith("</ifName>"):
            self.updates_cmd.pop()
            return ""

        xml_str += '</counter></counters>'

        return xml_str

    def config_export(self):
        """configure sflow export"""

        xml_str = ''
        if not self.export_route:
            return xml_str

        if self.export_route == "enable":
            if self.sflow_dict["export"] and self.sflow_dict["export"].get("ExportRoute") == "disable":
                xml_str = '<exports><export operation="delete"><ExportRoute>disable</ExportRoute></export></exports>'
                self.updates_cmd.append("undo sflow export extended-route-data disable")
        else:   # disable
            if not self.sflow_dict["export"] or self.sflow_dict["export"].get("ExportRoute") != "disable":
                xml_str = '<exports><export operation="create"><ExportRoute>disable</ExportRoute></export></exports>'
                self.updates_cmd.append("sflow export extended-route-data disable")

        return xml_str

    def netconf_load_config(self, xml_str):
        """load sflow config by netconf"""

        if not xml_str:
            return

        xml_cfg = """
            <config>
            <sflow xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
            %s
            </sflow>
            </config>""" % xml_str

        self.netconf_set_config(xml_cfg, "SET_SFLOW")
        self.changed = True

    def check_params(self):
        """Check all input params"""

        # check agent_ip
        if self.agent_ip:
            self.agent_ip = self.agent_ip.upper()
            if not check_ip_addr(self.agent_ip):
                self.module.fail_json(msg="Error: agent_ip is invalid.")

        # check source_ip
        if self.source_ip:
            self.source_ip = self.source_ip.upper()
            if not check_ip_addr(self.source_ip):
                self.module.fail_json(msg="Error: source_ip is invalid.")

        # check collector
        if self.collector_id:
            # check collector_ip and collector_ip_vpn
            if self.collector_ip:
                self.collector_ip = self.collector_ip.upper()
                if not check_ip_addr(self.collector_ip):
                    self.module.fail_json(
                        msg="Error: collector_ip is invalid.")
                if self.collector_ip_vpn and not is_valid_ip_vpn(self.collector_ip_vpn):
                    self.module.fail_json(
                        msg="Error: collector_ip_vpn is invalid.")

            # check collector_datagram_size ranges from 1024 to 8100
            if self.collector_datagram_size:
                if not self.collector_datagram_size.isdigit():
                    self.module.fail_json(
                        msg="Error: collector_datagram_size is not digit.")
                if int(self.collector_datagram_size) < 1024 or int(self.collector_datagram_size) > 8100:
                    self.module.fail_json(
                        msg="Error: collector_datagram_size is not ranges from 1024 to 8100.")

            # check collector_udp_port ranges from 1 to 65535
            if self.collector_udp_port:
                if not self.collector_udp_port.isdigit():
                    self.module.fail_json(
                        msg="Error: collector_udp_port is not digit.")
                if int(self.collector_udp_port) < 1 or int(self.collector_udp_port) > 65535:
                    self.module.fail_json(
                        msg="Error: collector_udp_port is not ranges from 1 to 65535.")

            # check collector_description 1 to 255 case-sensitive characters
            if self.collector_description:
                if self.collector_description.count(" "):
                    self.module.fail_json(
                        msg="Error: collector_description should without spaces.")
                if len(self.collector_description) < 1 or len(self.collector_description) > 255:
                    self.module.fail_json(
                        msg="Error: collector_description is not ranges from 1 to 255.")

        # check sflow_interface
        if self.sflow_interface:
            intf_type = get_interface_type(self.sflow_interface)
            if not intf_type:
                self.module.fail_json(msg="Error: intf_type is invalid.")
            if intf_type not in ['ge', '10ge', '25ge', '4x10ge', '40ge', '100ge', 'eth-trunk']:
                self.module.fail_json(
                    msg="Error: interface %s is not support sFlow." % self.sflow_interface)

            # check sample_collector
            if self.sample_collector:
                self.sample_collector.sort()
                if self.sample_collector not in [["1"], ["2"], ["1", "2"]]:
                    self.module.fail_json(
                        msg="Error: sample_collector is invalid.")

            # check sample_rate ranges from 1 to 4294967295
            if self.sample_rate:
                if not self.sample_rate.isdigit():
                    self.module.fail_json(
                        msg="Error: sample_rate is not digit.")
                if int(self.sample_rate) < 1 or int(self.sample_rate) > 4294967295:
                    self.module.fail_json(
                        msg="Error: sample_rate is not ranges from 1 to 4294967295.")

            # check sample_length ranges from 18 to 512
            if self.sample_length:
                if not self.sample_length.isdigit():
                    self.module.fail_json(
                        msg="Error: sample_rate is not digit.")
                if int(self.sample_length) < 18 or int(self.sample_length) > 512:
                    self.module.fail_json(
                        msg="Error: sample_length is not ranges from 18 to 512.")

            # check counter_collector
            if self.counter_collector:
                self.counter_collector.sort()
                if self.counter_collector not in [["1"], ["2"], ["1", "2"]]:
                    self.module.fail_json(
                        msg="Error: counter_collector is invalid.")

            # counter_interval ranges from 10 to 4294967295
            if self.counter_interval:
                if not self.counter_interval.isdigit():
                    self.module.fail_json(
                        msg="Error: counter_interval is not digit.")
                if int(self.counter_interval) < 10 or int(self.counter_interval) > 4294967295:
                    self.module.fail_json(
                        msg="Error: sample_length is not ranges from 10 to 4294967295.")

        if self.rate_limit or self.rate_limit_slot or self.forward_enp_slot:
            self.module.fail_json(msg="Error: The following parameters cannot be configured"
                                      "because XML mode is not supported:rate_limit,rate_limit_slot,forward_enp_slot.")

    def get_proposed(self):
        """get proposed info"""

        # base config
        if self.agent_ip:
            self.proposed["agent_ip"] = self.agent_ip
        if self.source_ip:
            self.proposed["source_ip"] = self.source_ip
        if self.export_route:
            self.proposed["export_route"] = self.export_route
        if self.rate_limit:
            self.proposed["rate_limit"] = self.rate_limit
            self.proposed["rate_limit_slot"] = self.rate_limit_slot
        if self.forward_enp_slot:
            self.proposed["forward_enp_slot"] = self.forward_enp_slot
        if self.collector_id:
            self.proposed["collector_id"] = self.collector_id
            if self.collector_ip:
                self.proposed["collector_ip"] = self.collector_ip
                self.proposed["collector_ip_vpn"] = self.collector_ip_vpn
            if self.collector_datagram_size:
                self.proposed[
                    "collector_datagram_size"] = self.collector_datagram_size
            if self.collector_udp_port:
                self.proposed["collector_udp_port"] = self.collector_udp_port
            if self.collector_meth:
                self.proposed["collector_meth"] = self.collector_meth
            if self.collector_description:
                self.proposed[
                    "collector_description"] = self.collector_description

        # sample and counter config
        if self.sflow_interface:
            self.proposed["sflow_interface"] = self.sflow_interface
            if self.sample_collector:
                self.proposed["sample_collector"] = self.sample_collector
            if self.sample_rate:
                self.proposed["sample_rate"] = self.sample_rate
            if self.sample_length:
                self.proposed["sample_length"] = self.sample_length
            if self.sample_direction:
                self.proposed["sample_direction"] = self.sample_direction
            if self.counter_collector:
                self.proposed["counter_collector"] = self.counter_collector
            if self.counter_interval:
                self.proposed["counter_interval"] = self.counter_interval

        self.proposed["state"] = self.state

    def get_existing(self):
        """get existing info"""

        if not self.sflow_dict:
            return

        if self.agent_ip:
            self.existing["agent"] = self.sflow_dict["agent"]
        if self.source_ip:
            self.existing["source"] = self.sflow_dict["source"]
        if self.collector_id:
            self.existing["collector"] = self.sflow_dict["collector"]
        if self.export_route:
            self.existing["export"] = self.sflow_dict["export"]

        if self.sflow_interface:
            self.existing["sampling"] = self.sflow_dict["sampling"]
            self.existing["counter"] = self.sflow_dict["counter"]

    def get_end_state(self):
        """get end state info"""

        sflow_dict = self.get_sflow_dict()
        if not sflow_dict:
            return

        if self.agent_ip:
            self.end_state["agent"] = sflow_dict["agent"]
        if self.source_ip:
            self.end_state["source"] = sflow_dict["source"]
        if self.collector_id:
            self.end_state["collector"] = sflow_dict["collector"]
        if self.export_route:
            self.end_state["export"] = sflow_dict["export"]

        if self.sflow_interface:
            self.end_state["sampling"] = sflow_dict["sampling"]
            self.end_state["counter"] = sflow_dict["counter"]
        if self.existing == self.end_state:
            self.changed = False

    def work(self):
        """worker"""

        self.check_params()
        self.sflow_dict = self.get_sflow_dict()
        self.get_existing()
        self.get_proposed()

        # deal present or absent
        xml_str = ''
        if self.export_route:
            xml_str += self.config_export()
        if self.agent_ip:
            xml_str += self.config_agent()
        if self.source_ip:
            xml_str += self.config_source()

        if self.state == "present":
            if self.collector_id and self.collector_ip:
                xml_str += self.config_collector()
            if self.sflow_interface:
                xml_str += self.config_sampling()
                xml_str += self.config_counter()
        else:
            if self.sflow_interface:
                xml_str += self.config_sampling()
                xml_str += self.config_counter()
            if self.collector_id:
                xml_str += self.config_collector()

        if xml_str:
            self.netconf_load_config(xml_str)
            self.changed = True

        self.get_end_state()
        self.results['changed'] = self.changed
        self.results['proposed'] = self.proposed
        self.results['existing'] = self.existing
        self.results['end_state'] = self.end_state
        if self.changed:
            self.results['updates'] = self.updates_cmd
        else:
            self.results['updates'] = list()

        self.module.exit_json(**self.results)


def main():
    """Module main"""

    argument_spec = dict(
        agent_ip=dict(required=False, type='str'),
        source_ip=dict(required=False, type='str'),
        export_route=dict(required=False, type='str',
                          choices=['enable', 'disable']),
        rate_limit=dict(required=False, removed_in_version='3.0.0',  # was Ansible 2.13
                        removed_from_collection='community.network', type='str'),
        rate_limit_slot=dict(required=False, removed_in_version='3.0.0',  # was Ansible 2.13
                             removed_from_collection='community.network', type='str'),
        forward_enp_slot=dict(required=False, removed_in_version='3.0.0',  # was Ansible 2.13
                              removed_from_collection='community.network', type='str'),
        collector_id=dict(required=False, type='str', choices=['1', '2']),
        collector_ip=dict(required=False, type='str'),
        collector_ip_vpn=dict(required=False, type='str'),
        collector_datagram_size=dict(required=False, type='str'),
        collector_udp_port=dict(required=False, type='str'),
        collector_meth=dict(required=False, type='str',
                            choices=['meth', 'enhanced']),
        collector_description=dict(required=False, type='str'),
        sflow_interface=dict(required=False, type='str'),
        sample_collector=dict(required=False, type='list'),
        sample_rate=dict(required=False, type='str'),
        sample_length=dict(required=False, type='str'),
        sample_direction=dict(required=False, type='str',
                              choices=['inbound', 'outbound', 'both']),
        counter_collector=dict(required=False, type='list'),
        counter_interval=dict(required=False, type='str'),
        state=dict(required=False, default='present',
                   choices=['present', 'absent'])
    )

    argument_spec.update(ce_argument_spec)
    module = Sflow(argument_spec)
    module.work()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Ansible, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_shell
short_description: Execute shell commands on target hosts
description:
     - The M(ansible.windows.win_shell) module takes the command name followed by a list of space-delimited arguments.
       It is similar to the M(ansible.windows.win_command) module, but runs
       the command via a shell (defaults to PowerShell) on the target host.
     - For non-Windows targets, use the M(ansible.builtin.shell) module instead.
options:
  free_form:
    description:
      - The M(ansible.windows.win_shell) module takes a free form command to run.
      - There is no parameter actually named 'free form'. See the examples!
    type: str
    required: yes
  creates:
    description:
      - A path or path filter pattern; when the referenced path exists on the target host, the task will be skipped.
    type: path
  removes:
    description:
      - A path or path filter pattern; when the referenced path B(does not) exist on the target host, the task will be skipped.
    type: path
  chdir:
    description:
      - Set the specified path as the current working directory before executing a command
    type: path
  executable:
    description:
      - Change the shell used to execute the command (eg, C(cmd)).
      - The target shell must accept a C(/c) parameter followed by the raw command line to be executed.
    type: path
  stdin:
    description:
    - Set the stdin of the command directly to the specified value.
    type: str
  no_profile:
    description:
    - Do not load the user profile before running a command. This is only valid
      when using PowerShell as the executable.
    type: bool
    default: no
  output_encoding_override:
    description:
    - This option overrides the encoding of stdout/stderr output.
    - You can use this option when you need to run a command which ignore the console's codepage.
    - You should only need to use this option in very rare circumstances.
    - This value can be any valid encoding C(Name) based on the output of C([System.Text.Encoding]::GetEncodings()).
      See U(https://docs.microsoft.com/dotnet/api/system.text.encoding.getencodings).
    type: str
notes:
   -  If you want to run an executable securely and predictably, it may be
      better to use the M(ansible.windows.win_command) module instead. Best practices when writing
      playbooks will follow the trend of using M(ansible.windows.win_command) unless C(win_shell) is
      explicitly required. When running ad-hoc commands, use your best judgement.
   -  WinRM will not return from a command execution until all child processes created have exited.
      Thus, it is not possible to use M(ansible.windows.win_shell) to spawn long-running child or background processes.
      Consider creating a Windows service for managing background processes.
seealso:
- module: community.windows.psexec
- module: ansible.builtin.raw
- module: ansible.builtin.script
- module: ansible.builtin.shell
- module: ansible.windows.win_command
- module: community.windows.win_psexec
author:
    - Matt Davis (@nitzmahone)
'''

EXAMPLES = r'''
- name: Execute a comand in the remote shell, stdout goes to the specified file on the remote
  ansible.windows.win_shell: C:\somescript.ps1 >> C:\somelog.txt

- name: Change the working directory to somedir/ before executing the command
  ansible.windows.win_shell: C:\somescript.ps1 >> C:\somelog.txt
  args:
    chdir: C:\somedir

- name: Run a command with an idempotent check on what it creates, will only run when somedir/somelog.txt does not exist
  ansible.windows.win_shell: C:\somescript.ps1 >> C:\somelog.txt
  args:
    chdir: C:\somedir
    creates: C:\somelog.txt

- name: Run a command under a non-Powershell interpreter (cmd in this case)
  ansible.windows.win_shell: echo %HOMEDIR%
  args:
    executable: cmd
  register: homedir_out

- name: Run multi-lined shell commands
  ansible.windows.win_shell: |
    $value = Test-Path -Path C:\temp
    if ($value) {
        Remove-Item -Path C:\temp -Force
    }
    New-Item -Path C:\temp -ItemType Directory

- name: Retrieve the input based on stdin
  ansible.windows.win_shell: '$string = [Console]::In.ReadToEnd(); Write-Output $string.Trim()'
  args:
    stdin: Input message
'''

RETURN = r'''
msg:
    description: Changed.
    returned: always
    type: bool
    sample: true
start:
    description: The command execution start time.
    returned: always
    type: str
    sample: '2016-02-25 09:18:26.429568'
end:
    description: The command execution end time.
    returned: always
    type: str
    sample: '2016-02-25 09:18:26.755339'
delta:
    description: The command execution delta time.
    returned: always
    type: str
    sample: '0:00:00.325771'
stdout:
    description: The command standard output.
    returned: always
    type: str
    sample: 'Clustering node rabbit@slave1 with rabbit@main ...'
stderr:
    description: The command standard error.
    returned: always
    type: str
    sample: 'ls: cannot access foo: No such file or directory'
cmd:
    description: The command executed by the task.
    returned: always
    type: str
    sample: 'rabbitmqctl join_cluster rabbit@main'
rc:
    description: The command return code (0 means success).
    returned: always
    type: int
    sample: 0
stdout_lines:
    description: The command standard output split in lines.
    returned: always
    type: list
    sample: [u'Clustering node rabbit@slave1 with rabbit@main ...']
'''



#!/usr/bin/python

# (c) 2017, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''

module: na_cdot_license

short_description: Manage NetApp cDOT protocol and feature licenses
extends_documentation_fragment:
- community.general._netapp.ontap

author: Sumit Kumar (@timuster) <sumit4@netapp.com>

deprecated:
  removed_in: 2.0.0  # was Ansible 2.11
  why: Updated modules released with increased functionality
  alternative: Use M(netapp.ontap.na_ontap_license) instead.

description:
- Add or remove licenses on NetApp ONTAP.

options:

  remove_unused:
    description:
    - Remove licenses that have no controller affiliation in the cluster.
    type: bool

  remove_expired:
    description:
    - Remove licenses that have expired in the cluster.
    type: bool

  serial_number:
    description:
    - Serial number of the node associated with the license.
    - This parameter is used primarily when removing license for a specific service.
    - If this parameter is not provided, the cluster serial number is used by default.

  licenses:
    description:
    - List of licenses to add or remove.
    - Please note that trying to remove a non-existent license will throw an error.
    suboptions:
      base:
        description:
          - Cluster Base License
      nfs:
        description:
          - NFS License
      cifs:
        description:
          - CIFS License
      iscsi:
        description:
          - iSCSI License
      fcp:
        description:
          - FCP License
      cdmi:
        description:
          - CDMI License
      snaprestore:
        description:
          - SnapRestore License
      snapmirror:
        description:
          - SnapMirror License
      flexclone:
        description:
          - FlexClone License
      snapvault:
        description:
          - SnapVault License
      snaplock:
        description:
          - SnapLock License
      snapmanagersuite:
        description:
          - SnapManagerSuite License
      snapprotectapps:
        description:
          - SnapProtectApp License
      v_storageattach:
        description:
          - Virtual Attached Storage License

'''


EXAMPLES = """
- name: Add licenses
  community.general.na_cdot_license:
    hostname: "{{ netapp_hostname }}"
    username: "{{ netapp_username }}"
    password: "{{ netapp_password }}"
    serial_number: #################
    licenses:
      nfs: #################
      cifs: #################
      iscsi: #################
      fcp: #################
      snaprestore: #################
      flexclone: #################

- name: Remove licenses
  community.general.na_cdot_license:
    hostname: "{{ netapp_hostname }}"
    username: "{{ netapp_username }}"
    password: "{{ netapp_password }}"
    remove_unused: false
    remove_expired: true
    serial_number: #################
    licenses:
      nfs: remove
"""

RETURN = """

"""
import traceback

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native
import ansible_collections.community.general.plugins.module_utils._netapp as netapp_utils


HAS_NETAPP_LIB = netapp_utils.has_netapp_lib()


class NetAppCDOTLicense(object):

    def __init__(self):
        self.argument_spec = netapp_utils.ontap_sf_host_argument_spec()
        self.argument_spec.update(dict(
            serial_number=dict(required=False, type='str', default=None),
            remove_unused=dict(default=False, type='bool'),
            remove_expired=dict(default=False, type='bool'),
            licenses=dict(default=False, type='dict'),
        ))

        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            supports_check_mode=False
        )

        p = self.module.params

        # set up state variables
        self.serial_number = p['serial_number']
        self.remove_unused = p['remove_unused']
        self.remove_expired = p['remove_expired']
        self.licenses = p['licenses']

        if HAS_NETAPP_LIB is False:
            self.module.fail_json(msg="the python NetApp-Lib module is required")
        else:
            self.server = netapp_utils.setup_ontap_zapi(module=self.module)

    def get_licensing_status(self):
        """
            Check licensing status

            :return: package (key) and licensing status (value)
            :rtype: dict
        """
        license_status = netapp_utils.zapi.NaElement('license-v2-status-list-info')
        result = None
        try:
            result = self.server.invoke_successfully(license_status,
                                                     enable_tunneling=False)
        except netapp_utils.zapi.NaApiError as e:
            self.module.fail_json(msg="Error checking license status: %s" %
                                  to_native(e), exception=traceback.format_exc())

        return_dictionary = {}
        license_v2_status = result.get_child_by_name('license-v2-status')
        if license_v2_status:
            for license_v2_status_info in license_v2_status.get_children():
                package = license_v2_status_info.get_child_content('package')
                status = license_v2_status_info.get_child_content('method')
                return_dictionary[package] = status

        return return_dictionary

    def remove_licenses(self, remove_list):
        """
        Remove requested licenses
        :param:
            remove_list : List of packages to remove

        """
        license_delete = netapp_utils.zapi.NaElement('license-v2-delete')
        for package in remove_list:
            license_delete.add_new_child('package', package)

        if self.serial_number is not None:
            license_delete.add_new_child('serial-number', self.serial_number)

        try:
            self.server.invoke_successfully(license_delete,
                                            enable_tunneling=False)
        except netapp_utils.zapi.NaApiError as e:
            self.module.fail_json(msg="Error removing license %s" %
                                  to_native(e), exception=traceback.format_exc())

    def remove_unused_licenses(self):
        """
        Remove unused licenses
        """
        remove_unused = netapp_utils.zapi.NaElement('license-v2-delete-unused')
        try:
            self.server.invoke_successfully(remove_unused,
                                            enable_tunneling=False)
        except netapp_utils.zapi.NaApiError as e:
            self.module.fail_json(msg="Error removing unused licenses: %s" %
                                  to_native(e), exception=traceback.format_exc())

    def remove_expired_licenses(self):
        """
        Remove expired licenses
        """
        remove_expired = netapp_utils.zapi.NaElement('license-v2-delete-expired')
        try:
            self.server.invoke_successfully(remove_expired,
                                            enable_tunneling=False)
        except netapp_utils.zapi.NaApiError as e:
            self.module.fail_json(msg="Error removing expired licenses: %s" %
                                  to_native(e), exception=traceback.format_exc())

    def update_licenses(self):
        """
        Update licenses
        """
        # Remove unused and expired licenses, if requested.
        if self.remove_unused:
            self.remove_unused_licenses()

        if self.remove_expired:
            self.remove_expired_licenses()

        # Next, add/remove specific requested licenses.
        license_add = netapp_utils.zapi.NaElement('license-v2-add')
        codes = netapp_utils.zapi.NaElement('codes')
        remove_list = []
        for key, value in self.licenses.items():
            str_value = str(value)
            # Make sure license is not an empty string.
            if str_value and str_value.strip():
                if str_value.lower() == 'remove':
                    remove_list.append(str(key).lower())
                else:
                    codes.add_new_child('license-code-v2', str_value)

        # Remove requested licenses.
        if len(remove_list) != 0:
            self.remove_licenses(remove_list)

        # Add requested licenses
        if len(codes.get_children()) != 0:
            license_add.add_child_elem(codes)
            try:
                self.server.invoke_successfully(license_add,
                                                enable_tunneling=False)
            except netapp_utils.zapi.NaApiError as e:
                self.module.fail_json(msg="Error adding licenses: %s" %
                                      to_native(e), exception=traceback.format_exc())

    def apply(self):
        changed = False
        # Add / Update licenses.
        license_status = self.get_licensing_status()
        self.update_licenses()
        new_license_status = self.get_licensing_status()

        if license_status != new_license_status:
            changed = True

        self.module.exit_json(changed=changed)


def main():
    v = NetAppCDOTLicense()
    v.apply()


if __name__ == '__main__':
    main()



# Copyright: (c) 2018, Pluribus Networks
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible_collections.community.network.tests.unit.compat.mock import patch
from ansible_collections.community.network.plugins.modules.network.netvisor import pn_port_config
from ansible_collections.community.network.tests.unit.plugins.modules.utils import set_module_args
from .nvos_module import TestNvosModule


class TestPortConfigModule(TestNvosModule):

    module = pn_port_config

    def setUp(self):
        self.mock_run_nvos_commands = patch('ansible_collections.community.network.plugins.modules.network.netvisor.pn_port_config.run_cli')
        self.run_nvos_commands = self.mock_run_nvos_commands.start()

    def tearDown(self):
        self.mock_run_nvos_commands.stop()

    def run_cli_patch(self, module, cli, state_map):
        if state_map['update'] == 'port-config-modify':
            results = dict(
                changed=True,
                cli_cmd=cli
            )
        module.exit_json(**results)

    def load_fixtures(self, commands=None, state=None, transport='cli'):
        self.run_nvos_commands.side_effect = self.run_cli_patch

    def test_pn_port_config_modify_t1(self):
        set_module_args({'pn_cliswitch': 'sw01', 'pn_port': '1,2',
                         'pn_speed': '10g', 'pn_jumbo': True, 'state': 'update'})
        result = self.execute_module(changed=True, state='update')
        expected_cmd = ' switch sw01 port-config-modify  speed 10g port 1,2 jumbo '
        self.assertEqual(result['cli_cmd'], expected_cmd)

    def test_pn_port_config_modify_t2(self):
        set_module_args({'pn_cliswitch': 'sw01', 'pn_port': 'all',
                         'pn_host_enable': True, 'state': 'update'})
        result = self.execute_module(changed=True, state='update')
        expected_cmd = ' switch sw01 port-config-modify  port all host-enable '
        self.assertEqual(result['cli_cmd'], expected_cmd)

    def test_pn_port_config_modify_t3(self):
        set_module_args({'pn_cliswitch': 'sw01', 'pn_port': '5',
                         'pn_crc_check_enable': True, 'pn_vxlan_termination': False, 'state': 'update'})
        result = self.execute_module(changed=True, state='update')
        expected_cmd = ' switch sw01 port-config-modify  port 5 crc-check-enable  no-vxlan-termination '
        self.assertEqual(result['cli_cmd'], expected_cmd)

    def test_pn_port_config_modify_t4(self):
        set_module_args({'pn_cliswitch': 'sw01', 'pn_port': '10,11,12',
                         'pn_pause': False, 'pn_enable': True, 'state': 'update'})
        result = self.execute_module(changed=True, state='update')
        expected_cmd = ' switch sw01 port-config-modify  port 10,11,12 no-pause  enable '
        self.assertEqual(result['cli_cmd'], expected_cmd)



#!/usr/bin/python
#
# Copyright (c) 2019 Zim Kalinowski, (@zikalino)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: azure_rm_virtualmachinescalesetinstance
version_added: "2.8"
short_description: Get Azure Virtual Machine Scale Set Instance facts
description:
    - Get facts of Azure Virtual Machine Scale Set VMs.

options:
    resource_group:
        description:
            - The name of the resource group.
        required: True
    vmss_name:
        description:
            - The name of the VM scale set.
        required: True
    instance_id:
        description:
            - The instance ID of the virtual machine.
        required: True
    latest_model:
        type: bool
        description:
            - Set to C(yes) to upgrade to the latest model.
    power_state:
        description:
            - Use this option to change power state of the instance.
        choices:
            - 'running'
            - 'stopped'
            - 'deallocated'
    protect_from_scale_in:
        type: bool
        description:
            - turn on/off instance protection from scale in
        version_added: "2.10"
    protect_from_scale_set_actions:
        type: bool
        description:
            - tun on/off instance protection from scale set actions
        version_added: "2.10"
    state:
        description:
            - State of the VMSS instance. Use C(present) to update an instance and C(absent) to delete an instance.
        default: present
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Zim Kalinowski (@zikalino)

'''

EXAMPLES = '''
  - name: Upgrade instance to the latest image
    azure_rm_virtualmachinescalesetinstance:
      resource_group: myResourceGroup
      vmss_name: myVMSS
      instance_id: "2"
      latest_model: yes

  - name: Turn on protect from scale in
    azure_rm_virtualmachinescalesetinstance:
        resource_group: myResourceGroup
        vmss_name: myVMSS
        instance_id: "2"
        protect_from_scale_in: true
'''

RETURN = '''
instances:
    description:
        - A list of instances.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Instance resource ID.
            returned: always
            type: str
            sample: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/TestGroup/providers/Microsoft.Compute/scalesets/myscaleset/vms/myvm
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from msrestazure.azure_exceptions import CloudError
    from azure.mgmt.compute import ComputeManagementClient
    from msrest.serialization import Model
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMVirtualMachineScaleSetInstance(AzureRMModuleBase):
    def __init__(self):
        # define user inputs into argument
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            vmss_name=dict(
                type='str',
                required=True
            ),
            instance_id=dict(
                type='str'
            ),
            latest_model=dict(
                type='bool'
            ),
            power_state=dict(
                type='str',
                choices=['running', 'stopped', 'deallocated']
            ),
            protect_from_scale_in=dict(
                type='bool'
            ),
            protect_from_scale_set_actions=dict(
                type='bool'
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )
        # store the results of the module operation
        self.results = dict(
            changed=False
        )
        self.mgmt_client = None
        self.resource_group = None
        self.vmss_name = None
        self.instance_id = None
        self.latest_model = None
        self.power_state = None
        self.state = None
        self.protect_from_scale_in = None
        self.protect_from_scale_set_actions = None
        super(AzureRMVirtualMachineScaleSetInstance, self).__init__(self.module_arg_spec, supports_tags=False)

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])
        self.mgmt_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                    base_url=self._cloud_environment.endpoints.resource_manager,
                                                    api_version='2019-07-01')

        instances = self.get()

        if self.state == 'absent':
            for item in instances:
                if not self.check_mode:
                    self.delete(item['instance_id'])
                self.results['changed'] = True
            self.results['instances'] = []
        else:
            if self.latest_model is not None:
                for item in instances:
                    if not item.get('latest_model', None):
                        if not self.check_mode:
                            self.apply_latest_model(item['instance_id'])
                        item['latest_model'] = True
                        self.results['changed'] = True

            if self.power_state is not None:
                for item in instances:
                    if self.power_state == 'stopped' and item['power_state'] not in ['stopped', 'stopping']:
                        if not self.check_mode:
                            self.stop(item['instance_id'])
                        self.results['changed'] = True
                    elif self.power_state == 'deallocated' and item['power_state'] not in ['deallocated']:
                        if not self.check_mode:
                            self.deallocate(item['instance_id'])
                        self.results['changed'] = True
                    elif self.power_state == 'running' and item['power_state'] not in ['running']:
                        if not self.check_mode:
                            self.start(item['instance_id'])
                        self.results['changed'] = True
            if self.protect_from_scale_in is not None or self.protect_from_scale_set_actions is not None:
                for item in instances:
                    protection_policy = item['protection_policy']
                    if protection_policy is None or self.protect_from_scale_in != protection_policy['protect_from_scale_in'] or \
                            self.protect_from_scale_set_actions != protection_policy['protect_from_scale_set_actions']:
                        if not self.check_mode:
                            self.update_protection_policy(self.instance_id, self.protect_from_scale_in, self.protect_from_scale_set_actions)
                        self.results['changed'] = True

        self.results['instances'] = [{'id': item['id']} for item in instances]
        return self.results

    def get(self):
        response = None
        results = []
        try:
            response = self.mgmt_client.virtual_machine_scale_set_vms.get(resource_group_name=self.resource_group,
                                                                          vm_scale_set_name=self.vmss_name,
                                                                          instance_id=self.instance_id)
            self.log("Response : {0}".format(response))
        except CloudError as e:
            self.log('Could not get facts for Virtual Machine Scale Set VM.')

        if response:
            results.append(self.format_response(response))

        return results

    def apply_latest_model(self, instance_id):
        try:
            poller = self.compute_client.virtual_machine_scale_sets.update_instances(resource_group_name=self.resource_group,
                                                                                     vm_scale_set_name=self.vmss_name,
                                                                                     instance_ids=[instance_id])
            self.get_poller_result(poller)
        except CloudError as exc:
            self.log("Error applying latest model {0} - {1}".format(self.vmss_name, str(exc)))
            self.fail("Error applying latest model {0} - {1}".format(self.vmss_name, str(exc)))

    def delete(self, instance_id):
        try:
            self.mgmt_client.virtual_machine_scale_set_vms.delete(resource_group_name=self.resource_group,
                                                                  vm_scale_set_name=self.vmss_name,
                                                                  instance_id=instance_id)
        except CloudError as e:
            self.log('Could not delete instance of Virtual Machine Scale Set VM.')
            self.fail('Could not delete instance of Virtual Machine Scale Set VM.')

    def start(self, instance_id):
        try:
            self.mgmt_client.virtual_machine_scale_set_vms.start(resource_group_name=self.resource_group,
                                                                 vm_scale_set_name=self.vmss_name,
                                                                 instance_id=instance_id)
        except CloudError as e:
            self.log('Could not start instance of Virtual Machine Scale Set VM.')
            self.fail('Could not start instance of Virtual Machine Scale Set VM.')

    def stop(self, instance_id):
        try:
            self.mgmt_client.virtual_machine_scale_set_vms.power_off(resource_group_name=self.resource_group,
                                                                     vm_scale_set_name=self.vmss_name,
                                                                     instance_id=instance_id)
        except CloudError as e:
            self.log('Could not stop instance of Virtual Machine Scale Set VM.')
            self.fail('Could not stop instance of Virtual Machine Scale Set VM.')

    def deallocate(self, instance_id):
        try:
            self.mgmt_client.virtual_machine_scale_set_vms.deallocate(resource_group_name=self.resource_group,
                                                                      vm_scale_set_name=self.vmss_name,
                                                                      instance_id=instance_id)
        except CloudError as e:
            self.log('Could not deallocate instance of Virtual Machine Scale Set VM.')
            self.fail('Could not deallocate instance of Virtual Machine Scale Set VM.')

    def update_protection_policy(self, instance_id, protect_from_scale_in, protect_from_scale_set_actions):
        try:
            d = {}
            if protect_from_scale_in is not None:
                d['protect_from_scale_in'] = protect_from_scale_in
            if protect_from_scale_set_actions is not None:
                d['protect_from_scale_set_actions'] = protect_from_scale_set_actions
            protection_policy = self.compute_models.VirtualMachineScaleSetVMProtectionPolicy(**d)
            instance = self.mgmt_client.virtual_machine_scale_set_vms.get(resource_group_name=self.resource_group,
                                                                          vm_scale_set_name=self.vmss_name,
                                                                          instance_id=instance_id)
            instance.protection_policy = protection_policy
            poller = self.mgmt_client.virtual_machine_scale_set_vms.update(resource_group_name=self.resource_group,
                                                                           vm_scale_set_name=self.vmss_name,
                                                                           instance_id=instance_id,
                                                                           parameters=instance)
            self.get_poller_result(poller)
        except CloudError as e:
            self.log('Could not update instance protection policy.')
            self.fail('Could not update instance protection policy.')

    def format_response(self, item):
        d = item.as_dict()
        iv = self.mgmt_client.virtual_machine_scale_set_vms.get_instance_view(resource_group_name=self.resource_group,
                                                                              vm_scale_set_name=self.vmss_name,
                                                                              instance_id=d.get('instance_id', None)).as_dict()
        power_state = ""
        for index in range(len(iv['statuses'])):
            code = iv['statuses'][index]['code'].split('/')
            if code[0] == 'PowerState':
                power_state = code[1]
                break
        d = {
            'id': d.get('id'),
            'tags': d.get('tags'),
            'instance_id': d.get('instance_id'),
            'latest_model': d.get('latest_model_applied'),
            'power_state': power_state,
            'protection_policy': d.get('protection_policy')
        }
        return d


def main():
    AzureRMVirtualMachineScaleSetInstance()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


class ModuleDocFragment(object):

    # Docker doc fragment
    DOCUMENTATION = r'''

options:
    docker_host:
        description:
            - The URL or Unix socket path used to connect to the Docker API. To connect to a remote host, provide the
              TCP connection string. For example, C(tcp://192.0.2.23:2376). If TLS is used to encrypt the connection,
              the module will automatically replace C(tcp) in the connection URL with C(https).
            - If the value is not specified in the task, the value of environment variable C(DOCKER_HOST) will be used
              instead. If the environment variable is not set, the default value will be used.
        type: str
        default: unix://var/run/docker.sock
        aliases: [ docker_url ]
    tls_hostname:
        description:
            - When verifying the authenticity of the Docker Host server, provide the expected name of the server.
            - If the value is not specified in the task, the value of environment variable C(DOCKER_TLS_HOSTNAME) will
              be used instead. If the environment variable is not set, the default value will be used.
        type: str
        default: localhost
    api_version:
        description:
            - The version of the Docker API running on the Docker Host.
            - Defaults to the latest version of the API supported by Docker SDK for Python and the docker daemon.
            - If the value is not specified in the task, the value of environment variable C(DOCKER_API_VERSION) will be
              used instead. If the environment variable is not set, the default value will be used.
        type: str
        default: auto
        aliases: [ docker_api_version ]
    timeout:
        description:
            - The maximum amount of time in seconds to wait on a response from the API.
            - If the value is not specified in the task, the value of environment variable C(DOCKER_TIMEOUT) will be used
              instead. If the environment variable is not set, the default value will be used.
        type: int
        default: 60
    ca_cert:
        description:
            - Use a CA certificate when performing server verification by providing the path to a CA certificate file.
            - If the value is not specified in the task and the environment variable C(DOCKER_CERT_PATH) is set,
              the file C(ca.pem) from the directory specified in the environment variable C(DOCKER_CERT_PATH) will be used.
        type: path
        aliases: [ tls_ca_cert, cacert_path ]
    client_cert:
        description:
            - Path to the client's TLS certificate file.
            - If the value is not specified in the task and the environment variable C(DOCKER_CERT_PATH) is set,
              the file C(cert.pem) from the directory specified in the environment variable C(DOCKER_CERT_PATH) will be used.
        type: path
        aliases: [ tls_client_cert, cert_path ]
    client_key:
        description:
            - Path to the client's TLS key file.
            - If the value is not specified in the task and the environment variable C(DOCKER_CERT_PATH) is set,
              the file C(key.pem) from the directory specified in the environment variable C(DOCKER_CERT_PATH) will be used.
        type: path
        aliases: [ tls_client_key, key_path ]
    ssl_version:
        description:
            - Provide a valid SSL version number. Default value determined by ssl.py module.
            - If the value is not specified in the task, the value of environment variable C(DOCKER_SSL_VERSION) will be
              used instead.
        type: str
    tls:
        description:
            - Secure the connection to the API by using TLS without verifying the authenticity of the Docker host
              server. Note that if I(validate_certs) is set to C(yes) as well, it will take precedence.
            - If the value is not specified in the task, the value of environment variable C(DOCKER_TLS) will be used
              instead. If the environment variable is not set, the default value will be used.
        type: bool
        default: no
    validate_certs:
        description:
            - Secure the connection to the API by using TLS and verifying the authenticity of the Docker host server.
            - If the value is not specified in the task, the value of environment variable C(DOCKER_TLS_VERIFY) will be
              used instead. If the environment variable is not set, the default value will be used.
        type: bool
        default: no
        aliases: [ tls_verify ]
    debug:
        description:
            - Debug mode
        type: bool
        default: no

notes:
  - Connect to the Docker daemon by providing parameters with each task or by defining environment variables.
    You can define C(DOCKER_HOST), C(DOCKER_TLS_HOSTNAME), C(DOCKER_API_VERSION), C(DOCKER_CERT_PATH), C(DOCKER_SSL_VERSION),
    C(DOCKER_TLS), C(DOCKER_TLS_VERIFY) and C(DOCKER_TIMEOUT). If you are using docker machine, run the script shipped
    with the product that sets up the environment. It will set these variables for you. See
    U(https://docs.docker.com/machine/reference/env/) for more details.
  - When connecting to Docker daemon with TLS, you might need to install additional Python packages.
    For the Docker SDK for Python, version 2.4 or newer, this can be done by installing C(docker[tls]) with M(ansible.builtin.pip).
  - Note that the Docker SDK for Python only allows to specify the path to the Docker configuration for very few functions.
    In general, it will use C($HOME/.docker/config.json) if the C(DOCKER_CONFIG) environment variable is not specified,
    and use C($DOCKER_CONFIG/config.json) otherwise.
'''

    # Additional, more specific stuff for minimal Docker SDK for Python version < 2.0

    DOCKER_PY_1_DOCUMENTATION = r'''
options: {}
requirements:
  - "Docker SDK for Python: Please note that the L(docker-py,https://pypi.org/project/docker-py/)
     Python module has been superseded by L(docker,https://pypi.org/project/docker/)
     (see L(here,https://github.com/docker/docker-py/issues/1310) for details).
     For Python 2.6, C(docker-py) must be used. Otherwise, it is recommended to
     install the C(docker) Python module. Note that both modules should *not*
     be installed at the same time. Also note that when both modules are installed
     and one of them is uninstalled, the other might no longer function and a
     reinstall of it is required."
'''

    # Additional, more specific stuff for minimal Docker SDK for Python version >= 2.0.
    # Note that Docker SDK for Python >= 2.0 requires Python 2.7 or newer.

    DOCKER_PY_2_DOCUMENTATION = r'''
options: {}
requirements:
  - "Python >= 2.7"
  - "Docker SDK for Python: Please note that the L(docker-py,https://pypi.org/project/docker-py/)
     Python module has been superseded by L(docker,https://pypi.org/project/docker/)
     (see L(here,https://github.com/docker/docker-py/issues/1310) for details).
     This module does *not* work with docker-py."
'''



# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible_collections.community.network.tests.unit.compat.mock import patch
from ansible_collections.community.network.plugins.modules.network.cnos import cnos_user
from ansible_collections.community.network.tests.unit.plugins.modules.utils import set_module_args
from .cnos_module import TestCnosModule, load_fixture


class TestCnosUserModule(TestCnosModule):

    module = cnos_user

    def setUp(self):
        super(TestCnosUserModule, self).setUp()
        self.mock_get_config = patch('ansible_collections.community.network.plugins.modules.network.cnos.cnos_user.get_config')
        self.get_config = self.mock_get_config.start()
        self.mock_load_config = patch('ansible_collections.community.network.plugins.modules.network.cnos.cnos_user.load_config')
        self.load_config = self.mock_load_config.start()
        self.mock_run_commands = patch('ansible_collections.community.network.plugins.modules.network.cnos.cnos_user.run_commands')
        self.run_commands = self.mock_run_commands.start()

    def tearDown(self):
        super(TestCnosUserModule, self).tearDown()

        self.mock_get_config.stop()
        self.mock_load_config.stop()
        self.mock_run_commands.stop()

    def load_fixtures(self, commands=None, transport='cli'):
        self.get_config.return_value = load_fixture('cnos_user_config.cfg')
        self.load_config.return_value = dict(diff=None, session='session')
        self.run_commands.return_value = [load_fixture('cnos_user_config.cfg')]

    def test_cnos_user_create(self):
        set_module_args(dict(name='test', configured_password='Anil'))
        commands = ['username test', 'username test password Anil']
        self.execute_module(changed=True, commands=commands)

    def test_cnos_user_delete(self):
        set_module_args(dict(name='ansible', state='absent'))
        commands = []
        self.execute_module(changed=False, commands=commands)

    def test_cnos_user_password(self):
        set_module_args(dict(name='ansible', configured_password='test'))
        commands = ['username ansible', 'username ansible password test']
        self.execute_module(changed=True, commands=commands)

    def test_cnos_user_purge(self):
        set_module_args(dict(purge=True))
        commands = ['no username admin\n', 'no username ansible\n']
        self.execute_module(changed=True, commands=commands)

    def test_cnos_user_role(self):
        set_module_args(dict(name='ansible', role='network-admin', configured_password='test'))
        result = self.execute_module(changed=True)
        self.assertIn('username ansible role network-admin', result['commands'])

    def test_cnos_user_sshkey(self):
        set_module_args(dict(name='ansible', sshkey='test'))
        commands = ['username ansible', 'username ansible sshkey test']
        self.execute_module(changed=True, commands=commands)

    def test_cnos_user_update_password_changed(self):
        set_module_args(dict(name='test', configured_password='test', update_password='on_create'))
        commands = ['username test', 'username test password test']
        self.execute_module(changed=True, commands=commands)

    def test_cnos_user_update_password_always(self):
        set_module_args(dict(name='ansible', configured_password='test', update_password='always'))
        commands = ['username ansible', 'username ansible password test']
        self.execute_module(changed=True, commands=commands)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, sky-joker
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
module: vmware_host_auto_start
short_description: Manage the auto power ON or OFF for vm on ESXi host
author:
  - sky-joker (@sky-joker)
description:
  - In this module, can set up automatic startup and shutdown of virtual machines according to host startup or shutdown.
requirements:
  - python >= 2.7
  - PyVmomi
options:
  esxi_hostname:
    description:
    - ESXi hostname where the VM to set auto power on or off exists.
    type: str
    required: True
  name:
    description:
    - VM name to set auto power on or off.
    - This is not necessary if change only system default VM settings for autoStart config.
    type: str
  uuid:
    description:
    - VM uuid to set auto power on or off, this is VMware's unique identifier.
    - This is required if C(name) is not supplied.
    - This is not necessary if change only system default VM settings for autoStart config.
    type: str
  use_instance_uuid:
    description:
    - Whether to use the VMware instance UUID rather than the BIOS UUID.
    type: bool
    default: no
  moid:
    description:
    - Managed Object ID of the instance to manage if known, this is a unique identifier only within a single vCenter instance.
    - This is required if C(name) or C(uuid) is not supplied.
    type: str
  system_defaults:
    description:
    - System defaults for auto-start or auto-stop config for virtual machine.
    type: dict
    suboptions:
      enabled:
        description:
        - Enable automatically start or stop of virtual machines.
        type: bool
        default: False
      start_delay:
        description:
        - Default auto start delay in seconds.
        type: int
        default: 120
      stop_action:
        description:
        - Default stop action executed on the virtual machine when the system stops.
        type: str
        choices: ['none', 'guestShutdown', 'powerOff', 'suspend']
        default: powerOff
      stop_delay:
        description:
        - Default auto stop delay in seconds.
        type: int
        default: 120
      wait_for_heartbeat:
        description:
        - Continue power on processing when VMware Tools started.
        - If this parameter is enabled to powers on the next virtual machine without waiting for the delay to pass.
        - However, the virtual machine must have VMware Tools installed.
        type: bool
        default: False
  power_info:
    description:
    - Startup or shutdown settings of virtual machine.
    - This setting will override the system defaults.
    type: dict
    default:
      start_action: none
      start_delay: -1
      start_order: -1
      stop_action: systemDefault
      stop_delay: -1
      wait_for_heartbeat: systemDefault
    suboptions:
      start_action:
        description:
        - Whether to start the virtual machine when the host startup.
        type: str
        choices: ['none', 'powerOn']
        default: none
      start_delay:
        description:
        - Auto start delay in seconds of virtual machine.
        type: int
        default: -1
      start_order:
        description:
        - The autostart priority of virtual machine.
        - Virtual machines with a lower number are powered on first.
        - On host shutdown, the virtual machines are shut down in reverse order, meaning those with a higher number are powered off first.
        type: int
        default: -1
      stop_action:
        description:
        - Stop action executed on the virtual machine when the system stops of virtual machine.
        choices: ['none', 'systemDefault', 'powerOff', 'suspend']
        type: str
        default: systemDefault
      stop_delay:
        description:
        - Auto stop delay in seconds of virtual machine.
        type: int
        default: -1
      wait_for_heartbeat:
        description:
        - Continue power on processing when VMware Tools started.
        type: str
        choices: ['no', 'yes', 'systemDefault']
        default: systemDefault
extends_documentation_fragment:
- community.vmware.vmware.documentation

'''

EXAMPLES = '''
---
- name: Update for system defaults config.
  community.vmware.vmware_host_auto_start:
    hostname: "{{ hostname }}"
    username: "{{ username }}"
    password: "{{ password }}"
    validate_certs: no
    esxi_hostname: "{{ esxi_hostname }}"
    system_defaults:
      enabled: yes
      start_delay: 100
      stop_action: guestShutdown

- name: Update for powerInfo config of virtual machine.
  community.vmware.vmware_host_auto_start:
    hostname: "{{ hostname }}"
    username: "{{ username }}"
    password: "{{ password }}"
    validate_certs: no
    esxi_hostname: "{{ esxi_hostname }}"
    name: "{{ vm_name }}"
    power_info:
      start_action: powerOn
      start_delay: 10
      start_order: 1
      stop_action: powerOff
      wait_for_heartbeat: yes
'''

RETURN = '''
system_defaults_config:
  description: Parameter return when system defaults config is changed.
  returned: changed
  type: dict
  sample: >-
    {
      "enabled": true,
      "start_delay": 120,
      "stop_action": "powerOff",
      "stop_delay": 120,
      "wait_for_heartbeat": false
    }
power_info_config:
  description: Parameter return when virtual machine power info config is changed.
  returned: changed
  type: dict
  sample: >-
    {
      "start_action": "powerOn",
      "start_delay": -1,
      "start_order": -1,
      "stop_action": "systemDefault",
      "stop_delay": -1,
      "wait_for_heartbeat": "systemDefault"
    }
'''

try:
    from pyVmomi import vim
except ImportError:
    pass

try:
    from collections import OrderedDict
except ImportError:
    try:
        from ordereddict import OrderedDict
    except ImportError:
        pass

from ansible.module_utils._text import to_native
from ansible_collections.community.vmware.plugins.module_utils.vmware import PyVmomi, vmware_argument_spec
from ansible.module_utils.basic import AnsibleModule


class VMwareHostAutoStartManager(PyVmomi):
    def __init__(self, module):
        super(VMwareHostAutoStartManager, self).__init__(module)
        self.esxi_hostname = self.params['esxi_hostname']
        self.name = self.params['name']
        self.uuid = self.params['uuid']
        self.moid = self.params['moid']
        self.system_defaults = self.params['system_defaults']
        self.power_info = self.params['power_info']

    def generate_system_defaults_config(self):
        system_defaults_config = vim.host.AutoStartManager.SystemDefaults()
        system_defaults_config.enabled = self.system_defaults['enabled']
        system_defaults_config.startDelay = self.system_defaults['start_delay']
        system_defaults_config.stopAction = self.system_defaults['stop_action']
        system_defaults_config.stopDelay = self.system_defaults['stop_delay']
        system_defaults_config.waitForHeartbeat = self.system_defaults['wait_for_heartbeat']

        return system_defaults_config

    def generate_power_info_config(self):
        power_info_config = vim.host.AutoStartManager.AutoPowerInfo()
        power_info_config.key = self.vm_obj
        power_info_config.startAction = self.power_info['start_action']
        power_info_config.startDelay = self.power_info['start_delay']
        power_info_config.startOrder = self.power_info['start_order']
        power_info_config.stopAction = self.power_info['stop_action']
        power_info_config.stopDelay = self.power_info['stop_delay']
        power_info_config.waitForHeartbeat = self.power_info['wait_for_heartbeat']

        return power_info_config

    def execute(self):
        result = dict(changed=False, diff={'before': {}, 'after': {}})

        host_obj = self.find_hostsystem_by_name(self.esxi_hostname)
        if not host_obj:
            self.module.fail_json(msg="Cannot find the specified ESXi host: %s" % self.esxi_hostname)

        self.vm_obj = None
        if self.name or self.uuid or self.moid:
            self.vm_obj = self.get_vm()
            if not self.vm_obj:
                self.module.fail_json(msg="Cannot find the specified VM: %s" % (self.name or self.uuid or self.moid))
            elif self.esxi_hostname != self.vm_obj.runtime.host.name:
                self.module.fail_json(msg="%s exists on another host: %s" % (self.name or self.uuid or self.moid, self.vm_obj.runtime.host.name))

        # Check the existing autoStart setting difference.
        system_defaults_config_difference = False
        existing_system_defaults = self.to_json(host_obj.config.autoStart.defaults)
        system_defaults_for_compare = dict(
            enabled=existing_system_defaults['enabled'],
            start_delay=existing_system_defaults['startDelay'],
            stop_action=existing_system_defaults['stopAction'],
            stop_delay=existing_system_defaults['stopDelay'],
            wait_for_heartbeat=existing_system_defaults['waitForHeartbeat']
        )
        if self.system_defaults:
            if 'guestshutdown' == system_defaults_for_compare['stop_action']:
                system_defaults_for_compare['stop_action'] = 'guestShutdown'

            if 'poweroff' == system_defaults_for_compare['stop_action']:
                system_defaults_for_compare['stop_action'] = 'powerOff'

            if system_defaults_for_compare != self.system_defaults:
                result['diff']['before']['system_defaults'] = OrderedDict(sorted(system_defaults_for_compare.items()))
                result['diff']['after']['system_defaults'] = OrderedDict(sorted(self.system_defaults.items()))
                system_defaults_config_difference = True

        # Check the existing autoStart powerInfo setting difference for VM.
        vm_power_info_config_difference = False
        existing_vm_power_info = {}
        if system_defaults_for_compare['enabled'] and self.vm_obj:
            for vm_power_info in host_obj.config.autoStart.powerInfo:
                if vm_power_info.key == self.vm_obj:
                    existing_vm_power_info = self.to_json(vm_power_info)
                    break

            if existing_vm_power_info:
                vm_power_info_for_compare = dict(
                    start_action=existing_vm_power_info['startAction'],
                    start_delay=existing_vm_power_info['startDelay'],
                    start_order=existing_vm_power_info['startOrder'],
                    stop_action=existing_vm_power_info['stopAction'],
                    stop_delay=existing_vm_power_info['stopDelay'],
                    wait_for_heartbeat=existing_vm_power_info['waitForHeartbeat']
                )
            else:
                vm_power_info_for_compare = dict(
                    start_action='none',
                    start_delay=-1,
                    start_order=-1,
                    stop_action='systemDefault',
                    stop_delay=-1,
                    wait_for_heartbeat='systemDefault'
                )

            if vm_power_info_for_compare != self.power_info:
                result['diff']['before']['power_info'] = OrderedDict(sorted(vm_power_info_for_compare.items()))
                result['diff']['after']['power_info'] = OrderedDict(sorted(self.power_info.items()))
                vm_power_info_config_difference = True

        auto_start_manager_config = vim.host.AutoStartManager.Config()
        auto_start_manager_config.powerInfo = []
        if system_defaults_config_difference or vm_power_info_config_difference:
            if system_defaults_config_difference:
                auto_start_manager_config.defaults = self.generate_system_defaults_config()
                result['system_defaults_config'] = self.system_defaults

            if vm_power_info_config_difference:
                auto_start_manager_config.powerInfo = [self.generate_power_info_config()]
                result['power_info_config'] = self.power_info

            if self.module.check_mode:
                result['changed'] = True
                self.module.exit_json(**result)

            try:
                host_obj.configManager.autoStartManager.ReconfigureAutostart(spec=auto_start_manager_config)
                result['changed'] = True
                self.module.exit_json(**result)
            except Exception as e:
                self.module.fail_json(msg=to_native(e))

            self.module.exit_json(**result)
        else:
            self.module.exit_json(**result)


def main():
    argument_spec = vmware_argument_spec()
    argument_spec.update(esxi_hostname=dict(type='str', required=True),
                         name=dict(type='str'),
                         uuid=dict(type='str'),
                         use_instance_uuid=dict(type='bool', default=False),
                         moid=dict(type='str'),
                         system_defaults=dict(type='dict',
                                              options=dict(
                                                  enabled=dict(type='bool', default=False),
                                                  start_delay=dict(type='int', default=120),
                                                  stop_action=dict(type='str', choices=['none', 'guestShutdown',
                                                                                        'powerOff', 'suspend'],
                                                                   default='powerOff'),
                                                  stop_delay=dict(type='int', default=120),
                                                  wait_for_heartbeat=dict(type='bool', default=False)),
                                              ),
                         power_info=dict(type='dict',
                                         options=dict(
                                             start_action=dict(type='str', choices=['none', 'powerOn'], default='none'),
                                             start_delay=dict(type='int', default=-1),
                                             start_order=dict(type='int', default=-1),
                                             stop_action=dict(type='str', choices=['none', 'systemDefault', 'powerOff',
                                                                                   'suspend'], default='systemDefault'),
                                             stop_delay=dict(type='int', default=-1),
                                             wait_for_heartbeat=dict(type='str', choices=['no', 'yes', 'systemDefault'],
                                                                     default='systemDefault')),
                                         default=dict(
                                             start_action='none',
                                             start_delay=-1,
                                             start_order=-1,
                                             stop_action='systemDefault',
                                             stop_delay=-1,
                                             wait_for_heartbeat='systemDefault'
                                         ))
                         )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    vmware_host_auto_start = VMwareHostAutoStartManager(module)
    vmware_host_auto_start.execute()


if __name__ == "__main__":
    main()



# Based on the chroot connection plugin by Maykel Moya
#
# (c) 2014, Lorin Hochstein
# (c) 2015, Leendert Brouwer (https://github.com/objectified)
# (c) 2015, Toshio Kuratomi <tkuratomi@ansible.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    author:
        - Lorin Hochestein
        - Leendert Brouwer
    connection: docker
    short_description: Run tasks in docker containers
    description:
        - Run commands or put/fetch files to an existing docker container.
    options:
      remote_user:
        description:
            - The user to execute as inside the container
        vars:
            - name: ansible_user
            - name: ansible_docker_user
      docker_extra_args:
        description:
            - Extra arguments to pass to the docker command line
        default: ''
      remote_addr:
        description:
            - The name of the container you want to access.
        default: inventory_hostname
        vars:
            - name: ansible_host
            - name: ansible_docker_host
'''

import distutils.spawn
import fcntl
import os
import os.path
import subprocess
import re

from distutils.version import LooseVersion

import ansible.constants as C
from ansible.compat import selectors
from ansible.errors import AnsibleError, AnsibleFileNotFound
from ansible.module_utils.six.moves import shlex_quote
from ansible.module_utils._text import to_bytes, to_native, to_text
from ansible.plugins.connection import ConnectionBase, BUFSIZE
from ansible.utils.display import Display

display = Display()


class Connection(ConnectionBase):
    ''' Local docker based connections '''

    transport = 'community.general.docker'
    has_pipelining = True

    def __init__(self, play_context, new_stdin, *args, **kwargs):
        super(Connection, self).__init__(play_context, new_stdin, *args, **kwargs)

        # Note: docker supports running as non-root in some configurations.
        # (For instance, setting the UNIX socket file to be readable and
        # writable by a specific UNIX group and then putting users into that
        # group).  Therefore we don't check that the user is root when using
        # this connection.  But if the user is getting a permission denied
        # error it probably means that docker on their system is only
        # configured to be connected to by root and they are not running as
        # root.

        # Windows uses Powershell modules
        if getattr(self._shell, "_IS_WINDOWS", False):
            self.module_implementation_preferences = ('.ps1', '.exe', '')

        if 'docker_command' in kwargs:
            self.docker_cmd = kwargs['docker_command']
        else:
            self.docker_cmd = distutils.spawn.find_executable('docker')
            if not self.docker_cmd:
                raise AnsibleError("docker command not found in PATH")

        docker_version = self._get_docker_version()
        if docker_version == u'dev':
            display.warning(u'Docker version number is "dev". Will assume latest version.')
        if docker_version != u'dev' and LooseVersion(docker_version) < LooseVersion(u'1.3'):
            raise AnsibleError('docker connection type requires docker 1.3 or higher')

        # The remote user we will request from docker (if supported)
        self.remote_user = None
        # The actual user which will execute commands in docker (if known)
        self.actual_user = None

        if self._play_context.remote_user is not None:
            if docker_version == u'dev' or LooseVersion(docker_version) >= LooseVersion(u'1.7'):
                # Support for specifying the exec user was added in docker 1.7
                self.remote_user = self._play_context.remote_user
                self.actual_user = self.remote_user
            else:
                self.actual_user = self._get_docker_remote_user()

                if self.actual_user != self._play_context.remote_user:
                    display.warning(u'docker {0} does not support remote_user, using container default: {1}'
                                    .format(docker_version, self.actual_user or u'?'))
        elif self._display.verbosity > 2:
            # Since we're not setting the actual_user, look it up so we have it for logging later
            # Only do this if display verbosity is high enough that we'll need the value
            # This saves overhead from calling into docker when we don't need to
            self.actual_user = self._get_docker_remote_user()

    @staticmethod
    def _sanitize_version(version):
        return re.sub(u'[^0-9a-zA-Z.]', u'', version)

    def _old_docker_version(self):
        cmd_args = []
        if self._play_context.docker_extra_args:
            cmd_args += self._play_context.docker_extra_args.split(' ')

        old_version_subcommand = ['version']

        old_docker_cmd = [self.docker_cmd] + cmd_args + old_version_subcommand
        p = subprocess.Popen(old_docker_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        cmd_output, err = p.communicate()

        return old_docker_cmd, to_native(cmd_output), err, p.returncode

    def _new_docker_version(self):
        # no result yet, must be newer Docker version
        cmd_args = []
        if self._play_context.docker_extra_args:
            cmd_args += self._play_context.docker_extra_args.split(' ')

        new_version_subcommand = ['version', '--format', "'{{.Server.Version}}'"]

        new_docker_cmd = [self.docker_cmd] + cmd_args + new_version_subcommand
        p = subprocess.Popen(new_docker_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        cmd_output, err = p.communicate()
        return new_docker_cmd, to_native(cmd_output), err, p.returncode

    def _get_docker_version(self):

        cmd, cmd_output, err, returncode = self._old_docker_version()
        if returncode == 0:
            for line in to_text(cmd_output, errors='surrogate_or_strict').split(u'\n'):
                if line.startswith(u'Server version:'):  # old docker versions
                    return self._sanitize_version(line.split()[2])

        cmd, cmd_output, err, returncode = self._new_docker_version()
        if returncode:
            raise AnsibleError('Docker version check (%s) failed: %s' % (to_native(cmd), to_native(err)))

        return self._sanitize_version(to_text(cmd_output, errors='surrogate_or_strict'))

    def _get_docker_remote_user(self):
        """ Get the default user configured in the docker container """
        p = subprocess.Popen([self.docker_cmd, 'inspect', '--format', '{{.Config.User}}', self._play_context.remote_addr],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        out, err = p.communicate()
        out = to_text(out, errors='surrogate_or_strict')

        if p.returncode != 0:
            display.warning(u'unable to retrieve default user from docker container: %s %s' % (out, to_text(err)))
            return None

        # The default exec user is root, unless it was changed in the Dockerfile with USER
        return out.strip() or u'root'

    def _build_exec_cmd(self, cmd):
        """ Build the local docker exec command to run cmd on remote_host

            If remote_user is available and is supported by the docker
            version we are using, it will be provided to docker exec.
        """

        local_cmd = [self.docker_cmd]

        if self._play_context.docker_extra_args:
            local_cmd += self._play_context.docker_extra_args.split(' ')

        local_cmd += [b'exec']

        if self.remote_user is not None:
            local_cmd += [b'-u', self.remote_user]

        # -i is needed to keep stdin open which allows pipelining to work
        local_cmd += [b'-i', self._play_context.remote_addr] + cmd

        return local_cmd

    def _connect(self, port=None):
        """ Connect to the container. Nothing to do """
        super(Connection, self)._connect()
        if not self._connected:
            display.vvv(u"ESTABLISH DOCKER CONNECTION FOR USER: {0}".format(
                self.actual_user or u'?'), host=self._play_context.remote_addr
            )
            self._connected = True

    def exec_command(self, cmd, in_data=None, sudoable=False):
        """ Run a command on the docker host """
        super(Connection, self).exec_command(cmd, in_data=in_data, sudoable=sudoable)

        local_cmd = self._build_exec_cmd([self._play_context.executable, '-c', cmd])

        display.vvv(u"EXEC {0}".format(to_text(local_cmd)), host=self._play_context.remote_addr)
        display.debug("opening command with Popen()")

        local_cmd = [to_bytes(i, errors='surrogate_or_strict') for i in local_cmd]

        p = subprocess.Popen(
            local_cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        display.debug("done running command with Popen()")

        if self.become and self.become.expect_prompt() and sudoable:
            fcntl.fcntl(p.stdout, fcntl.F_SETFL, fcntl.fcntl(p.stdout, fcntl.F_GETFL) | os.O_NONBLOCK)
            fcntl.fcntl(p.stderr, fcntl.F_SETFL, fcntl.fcntl(p.stderr, fcntl.F_GETFL) | os.O_NONBLOCK)
            selector = selectors.DefaultSelector()
            selector.register(p.stdout, selectors.EVENT_READ)
            selector.register(p.stderr, selectors.EVENT_READ)

            become_output = b''
            try:
                while not self.become.check_success(become_output) and not self.become.check_password_prompt(become_output):
                    events = selector.select(self._play_context.timeout)
                    if not events:
                        stdout, stderr = p.communicate()
                        raise AnsibleError('timeout waiting for privilege escalation password prompt:\n' + to_native(become_output))

                    for key, event in events:
                        if key.fileobj == p.stdout:
                            chunk = p.stdout.read()
                        elif key.fileobj == p.stderr:
                            chunk = p.stderr.read()

                    if not chunk:
                        stdout, stderr = p.communicate()
                        raise AnsibleError('privilege output closed while waiting for password prompt:\n' + to_native(become_output))
                    become_output += chunk
            finally:
                selector.close()

            if not self.become.check_success(become_output):
                become_pass = self.become.get_option('become_pass', playcontext=self._play_context)
                p.stdin.write(to_bytes(become_pass, errors='surrogate_or_strict') + b'\n')
            fcntl.fcntl(p.stdout, fcntl.F_SETFL, fcntl.fcntl(p.stdout, fcntl.F_GETFL) & ~os.O_NONBLOCK)
            fcntl.fcntl(p.stderr, fcntl.F_SETFL, fcntl.fcntl(p.stderr, fcntl.F_GETFL) & ~os.O_NONBLOCK)

        display.debug("getting output with communicate()")
        stdout, stderr = p.communicate(in_data)
        display.debug("done communicating")

        display.debug("done with docker.exec_command()")
        return (p.returncode, stdout, stderr)

    def _prefix_login_path(self, remote_path):
        ''' Make sure that we put files into a standard path

            If a path is relative, then we need to choose where to put it.
            ssh chooses $HOME but we aren't guaranteed that a home dir will
            exist in any given chroot.  So for now we're choosing "/" instead.
            This also happens to be the former default.

            Can revisit using $HOME instead if it's a problem
        '''
        if getattr(self._shell, "_IS_WINDOWS", False):
            import ntpath
            return ntpath.normpath(remote_path)
        else:
            if not remote_path.startswith(os.path.sep):
                remote_path = os.path.join(os.path.sep, remote_path)
            return os.path.normpath(remote_path)

    def put_file(self, in_path, out_path):
        """ Transfer a file from local to docker container """
        super(Connection, self).put_file(in_path, out_path)
        display.vvv("PUT %s TO %s" % (in_path, out_path), host=self._play_context.remote_addr)

        out_path = self._prefix_login_path(out_path)
        if not os.path.exists(to_bytes(in_path, errors='surrogate_or_strict')):
            raise AnsibleFileNotFound(
                "file or module does not exist: %s" % to_native(in_path))

        out_path = shlex_quote(out_path)
        # Older docker doesn't have native support for copying files into
        # running containers, so we use docker exec to implement this
        # Although docker version 1.8 and later provide support, the
        # owner and group of the files are always set to root
        with open(to_bytes(in_path, errors='surrogate_or_strict'), 'rb') as in_file:
            if not os.fstat(in_file.fileno()).st_size:
                count = ' count=0'
            else:
                count = ''
            args = self._build_exec_cmd([self._play_context.executable, "-c", "dd of=%s bs=%s%s" % (out_path, BUFSIZE, count)])
            args = [to_bytes(i, errors='surrogate_or_strict') for i in args]
            try:
                p = subprocess.Popen(args, stdin=in_file,
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            except OSError:
                raise AnsibleError("docker connection requires dd command in the container to put files")
            stdout, stderr = p.communicate()

            if p.returncode != 0:
                raise AnsibleError("failed to transfer file %s to %s:\n%s\n%s" %
                                   (to_native(in_path), to_native(out_path), to_native(stdout), to_native(stderr)))

    def fetch_file(self, in_path, out_path):
        """ Fetch a file from container to local. """
        super(Connection, self).fetch_file(in_path, out_path)
        display.vvv("FETCH %s TO %s" % (in_path, out_path), host=self._play_context.remote_addr)

        in_path = self._prefix_login_path(in_path)
        # out_path is the final file path, but docker takes a directory, not a
        # file path
        out_dir = os.path.dirname(out_path)

        args = [self.docker_cmd, "cp", "%s:%s" % (self._play_context.remote_addr, in_path), out_dir]
        args = [to_bytes(i, errors='surrogate_or_strict') for i in args]

        p = subprocess.Popen(args, stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate()

        if getattr(self._shell, "_IS_WINDOWS", False):
            import ntpath
            actual_out_path = ntpath.join(out_dir, ntpath.basename(in_path))
        else:
            actual_out_path = os.path.join(out_dir, os.path.basename(in_path))

        if p.returncode != 0:
            # Older docker doesn't have native support for fetching files command `cp`
            # If `cp` fails, try to use `dd` instead
            args = self._build_exec_cmd([self._play_context.executable, "-c", "dd if=%s bs=%s" % (in_path, BUFSIZE)])
            args = [to_bytes(i, errors='surrogate_or_strict') for i in args]
            with open(to_bytes(actual_out_path, errors='surrogate_or_strict'), 'wb') as out_file:
                try:
                    p = subprocess.Popen(args, stdin=subprocess.PIPE,
                                         stdout=out_file, stderr=subprocess.PIPE)
                except OSError:
                    raise AnsibleError("docker connection requires dd command in the container to put files")
                stdout, stderr = p.communicate()

                if p.returncode != 0:
                    raise AnsibleError("failed to fetch file %s to %s:\n%s\n%s" % (in_path, out_path, stdout, stderr))

        # Rename if needed
        if actual_out_path != out_path:
            os.rename(to_bytes(actual_out_path, errors='strict'), to_bytes(out_path, errors='strict'))

    def close(self):
        """ Terminate the connection. Nothing to do for Docker"""
        super(Connection, self).close()
        self._connected = False
